from clayrs.recsys.graphs.nx_implementation import NXFullGraph
import os

from clayrs.recsys.graphs.graph import PropertyNode, UserNode
from clayrs.utils import load_content_instance
from test.recsys.graphs.test_networkx_implementation.test_nx_tripartite_graphs import TestNXTripartiteGraph
from test.recsys.graphs.test_networkx_implementation.test_nx_bipartite_graphs import rat, rat_timestamp
from test import dir_test_files

ratings_filename = os.path.join(dir_test_files, 'new_ratings_small.csv')
movies_dir = os.path.join(dir_test_files, 'complex_contents', 'movies_codified/')
users_dir = os.path.join(dir_test_files, 'complex_contents', 'users_codified/')


class TestNXFullGraph(TestNXTripartiteGraph):

    def setUp(self) -> None:
        # graphs that will be used for testing
        self.g: NXFullGraph = NXFullGraph(rat,
                                          item_contents_dir=movies_dir,
                                          item_exo_properties={'dbpedia': ['film director',
                                                                           'runtime (m)']},

                                          # It's the column in the users .DAT which identifies the gender
                                          user_exo_properties={'local': '1'},
                                          user_contents_dir=users_dir)

        self.graph_custom_label: NXFullGraph = NXFullGraph(rat, link_label='my_label',
                                                           item_contents_dir=movies_dir,
                                                           item_exo_properties={'dbpedia': ['film director',
                                                                                            'runtime (m)']},

                                                           # It's the column in the users .DAT which identifies the
                                                           # gender
                                                           user_exo_properties={'local': '1'},
                                                           user_contents_dir=users_dir)

        self.graph_timestamp: NXFullGraph = NXFullGraph(rat_timestamp,
                                                        item_contents_dir=movies_dir,
                                                        item_exo_properties={'dbpedia': ['film director',
                                                                                         'runtime (m)']},

                                                        # It's the column in the users .DAT which identifies the gender
                                                        user_exo_properties={'local': '1'},
                                                        user_contents_dir=users_dir)

        # this will be empty even if other attributes are specified since ratings are missing
        self.empty_graph: NXFullGraph = NXFullGraph(item_contents_dir=movies_dir,
                                                    item_exo_properties={'dbpedia': ['film director',
                                                                                     'runtime (m)']},

                                                    # It's the column in the users .DAT which identifies the gender
                                                    user_exo_properties={'local': '1'},
                                                    user_contents_dir=users_dir)

        # item_exo_properties set but no item_contents_dir specified
        self.graph_missing_item_dir_parameter = NXFullGraph(rat, item_exo_properties={'dbpedia': ['film director',
                                                                                                  'runtime (m)']})

        # item_contents_dir set but no item_exo_properties_specified specified
        self.graph_missing_item_prop_parameter = NXFullGraph(rat, item_contents_dir=movies_dir)

        # user_exo_properties set but no user_contents_dir specified
        self.graph_missing_user_dir_parameter = NXFullGraph(rat, user_exo_properties={'local'})

        # user_contents_dir set but no user_exo_properties_specified specified
        self.graph_missing_user_prop_parameter = NXFullGraph(rat, user_contents_dir=users_dir)

    def test_graph_creation(self):
        # the super class test will check if every user and item have a link
        # as they are present in the ratings frame and that each item is linked to its property
        super().test_graph_creation()

        # here we test if item nodes are linked to their exogenous property as specified in the constructor
        for user_node in self.g.user_nodes:
            loaded_item = load_content_instance(users_dir, user_node.value)
            exogenous_representation: dict = loaded_item.get_exogenous_representation("local").value
            gender_prop_expected = exogenous_representation.get("1", [])

            if not isinstance(gender_prop_expected, list):
                gender_prop_expected = [gender_prop_expected]

            for gender in gender_prop_expected:
                self.assertTrue(PropertyNode(gender) in self.g.property_nodes)
                result_link_data = self.g.get_link_data(user_node, PropertyNode(gender))
                expected_link_data = {'label': '1'}

                self.assertEqual(expected_link_data, result_link_data)

    def test_graph_creation_missing_parameter(self):

        # super class method will test missing parameters related to items
        super().test_graph_creation_missing_parameter()

        self.assertTrue(len(self.graph_missing_user_dir_parameter.user_nodes) != 0)
        self.assertTrue(len(self.graph_missing_user_dir_parameter.item_nodes) != 0)
        # warning is printed and no prop will be loaded
        self.assertTrue(len(self.graph_missing_user_dir_parameter.property_nodes) == 0)

        self.assertTrue(len(self.graph_missing_user_prop_parameter.user_nodes) != 0)
        self.assertTrue(len(self.graph_missing_user_prop_parameter.item_nodes) != 0)
        # warning is printed and no prop will be loaded
        self.assertTrue(len(self.graph_missing_user_prop_parameter.property_nodes) == 0)

    def test_add_link_user_prop_existent(self):
        # Link existent 'user' node to an existent 'prop' node
        user_node = UserNode('u_prop')
        prop_node = PropertyNode('Nolan')
        self.g.add_node(user_node)
        self.g.add_node(prop_node)
        self.assertIsNone(self.g.get_link_data(user_node, prop_node))
        self.g.add_link(user_node, prop_node, timestamp='now', label='directed_by')
        expected = {'timestamp': 'now', 'label': 'directed_by'}
        result = self.g.get_link_data(user_node, prop_node)
        self.assertEqual(expected, result)

        # Link list of existent 'user' nodes to a list of existent 'property' nodes
        items_list = [UserNode('i1_list'), UserNode('i2_list'), UserNode('i3_list')]
        self.g.add_node(items_list)
        props_list = [PropertyNode('p1_list'), PropertyNode('p2_list'), PropertyNode('p3_list')]
        self.g.add_node(props_list)
        self.g.add_link(items_list, props_list, 0.5)
        for user, item in zip(items_list, props_list):
            result = self.g.get_link_data(user, item)
            expected = {'weight': 0.5}

            self.assertEqual(expected, result)

        # Link existent 'property' node to an existent 'item' node
        user_node = UserNode('u_prop')
        prop_node = PropertyNode('Nolan')
        self.g.add_node(user_node)
        self.g.add_node(prop_node)
        self.assertIsNone(self.g.get_link_data(prop_node, user_node))
        self.g.add_link(prop_node, user_node)
        self.assertIsNotNone(self.g.get_link_data(user_node, prop_node))

        # Link list of existent 'prop' nodes to a list of existent 'item' nodes
        items_list = [UserNode('i1_list'), UserNode('i2_list'), UserNode('i3_list')]
        self.g.add_node(items_list)
        users_list = [UserNode('u1_list'), UserNode('u2_list'), UserNode('u3_list')]
        self.g.add_node(users_list)
        self.g.add_link(items_list, users_list, 0.5)
        for item, user in zip(items_list, users_list):
            result = self.g.get_link_data(item, user)
            expected = {'weight': 0.5}

            self.assertEqual(expected, result)

    def test_add_link_user_prop_non_existent(self):
        # Link non-existent 'user' node and non-existent 'item' node,
        # so both nodes are created and then linked
        user_new = UserNode('u_new')
        prop_new = PropertyNode('p_new')
        self.assertFalse(self.g.node_exists(user_new))
        self.assertFalse(self.g.node_exists(prop_new))
        self.g.add_link(user_new, prop_new, 0.5)
        self.assertTrue(self.g.node_exists(user_new))
        self.assertTrue(self.g.node_exists(prop_new))
        self.assertIsNotNone(self.g.get_link_data(user_new, prop_new))

        # Link non-existent 'user' node list and non-existent 'item' node list,
        # so all nodes of the two lists are created and then linked
        user_new_list = [UserNode('i_new_new1'), UserNode('i_new_new2')]
        prop_new_list = [PropertyNode('u_new_new1'), PropertyNode('u_new_new2')]
        for user in user_new_list:
            self.assertFalse(self.g.node_exists(user))
        for prop in prop_new_list:
            self.assertFalse(self.g.node_exists(prop))

        self.g.add_link(user_new_list, prop_new_list, 0.5)

        for user, item in zip(user_new_list, prop_new_list):
            result = self.g.get_link_data(user, item)
            expected = {'weight': 0.5}

            self.assertEqual(expected, result)

    # this needs to override superclass test, since for tripartite graph we have no possible error
    # when adding new links
    def test_add_link_raise_error(self):
        pass

    def test_add_node_with_prop(self):
        super().test_add_node_with_prop()

        # add 'user' node and only properties '2' to the graph
        self.assertFalse(UserNode('6') in self.g.user_nodes)
        self.assertFalse(PropertyNode('50') in self.g.property_nodes)
        self.g.add_node_with_prop(UserNode('6'), {'local': ['2']}, users_dir)
        self.assertTrue(UserNode('6') in self.g.user_nodes)
        self.assertTrue(PropertyNode('50') in self.g.property_nodes)
        self.assertIsNotNone(self.g.get_link_data(UserNode('6'),
                                                  PropertyNode('50')))

        self.g.remove_node(UserNode('6'))
        self.g.remove_node(PropertyNode('50'))

        # add 'user' node and only property'2' to the graph with different id from the file name
        self.assertFalse(UserNode('different') in self.g.user_nodes)
        self.assertFalse(PropertyNode('50') in self.g.property_nodes)
        self.g.add_node_with_prop(UserNode('different'), {'local': ['2']}, users_dir,
                                  '6')
        self.assertTrue(UserNode('different') in self.g.user_nodes)
        self.assertTrue(PropertyNode('50') in self.g.property_nodes)
        self.assertIsNotNone(self.g.get_link_data(UserNode('different'),
                                                  PropertyNode('50')))

        self.g.remove_node(UserNode('different'))
        self.g.remove_node(PropertyNode('50'))

        # add 'item' node and all its properties
        loaded_item = load_content_instance(users_dir, '6')
        exogenous_representation: dict = loaded_item.get_exogenous_representation("local").value

        self.assertFalse(UserNode('6') in self.g.item_nodes)
        self.assertFalse(PropertyNode('50') in self.g.property_nodes)
        self.g.add_node_with_prop(UserNode('6'), {'local'}, users_dir)
        self.assertTrue(UserNode('6') in self.g.user_nodes)

        for prop_label in exogenous_representation.keys():
            prop_val_expected = exogenous_representation.get(prop_label, [])

            if not isinstance(prop_val_expected, list):
                prop_val_expected = [prop_val_expected]

            for prop_val in prop_val_expected:
                self.assertTrue(PropertyNode(prop_val) in self.g.property_nodes)
                result_link_data_list = self.g.get_link_data(UserNode('6'), PropertyNode(prop_val))

                # we make sure that this is a property of the exogenous representation
                self.assertIn(result_link_data_list['label'], exogenous_representation)

    # this needs to override superclass test, since for tripartite graph we have no possible error
    # when adding new nodes
    def test_add_raise_error(self):
        pass
    # def test_convert_to_dataframe(self):
    #     converted_df = self.g.convert_to_dataframe()
    #     self.assertNotIn('label', converted_df.columns)
    #     for user, item in zip(converted_df['from_id'], converted_df['to_id']):
    #         self.assertIsInstance(user, Node)
    #         self.assertIsInstance(item, Node)
    #
    #     converted_df = converted_df.query('to_id not in @self.g.property_nodes')
    #     result = np.sort(converted_df, axis=0)
    #     expected = np.sort(self.df, axis=0)
    #     self.assertTrue(np.array_equal(expected, result))
    #
    #     converted_df = self.g.convert_to_dataframe(only_values=True, with_label=True)
    #     self.assertIn('label', converted_df.columns)
    #     for user, item in zip(converted_df['from_id'], converted_df['to_id']):
    #         self.assertNotIsInstance(user, Node)
    #         self.assertNotIsInstance(item, Node)
    #
    #     converted_df = converted_df.query('to_id not in @self.g.property_nodes')[['from_id', 'to_id', 'score']]
    #     result = np.sort(converted_df, axis=0)
    #     expected = np.sort(self.df, axis=0)
    #     self.assertTrue(np.array_equal(expected, result))



import unittest
from unittest import TestCase

import pandas as pd
import os

from clayrs.content_analyzer import Ratings
from clayrs.recsys.content_based_algorithm.classifier.classifier_recommender import ClassifierRecommender
from clayrs.recsys.content_based_algorithm.classifier.classifiers import Classifier,\
    SkGaussianProcess, SkRandomForest, SkLogisticRegression, SkKNN, SkSVC, SkDecisionTree
from clayrs.recsys.content_based_algorithm.contents_loader import LoadedContentsDict
from clayrs.recsys.content_based_algorithm.exceptions import OnlyNegativeItems, NoRatedItems, \
    OnlyPositiveItems, NotPredictionAlg, EmptyUserRatings
from test import dir_test_files


def for_each_classifier(test_func):
    def wrapper(self, *args, **kwargs):
        for classifier in self.classifiers_list:
            with self.subTest(current_classifier=classifier):
                test_func(*((self, classifier) + args), **kwargs)

    return wrapper


class TestClassifierRecommender(TestCase):

    def setUp(self) -> None:
        ratings = pd.DataFrame.from_records([
            ("A000", "tt0112281", 0.5, "54654675"),
            ("A000", "tt0112302", 0.5, "54654675"),
            ("A001", "tt0114576", 0.8, "54654675"),
            ("A001", "tt0112896", -0.4, "54654675"),
            ("A000", "tt0112346", -0.5, "54654675"),
            ("A000", "tt0112453", -0.5, "54654675"),
            ("A002", "tt0112453", -0.2, "54654675"),
            ("A002", "tt0113497", 0.5, "54654675"),
            ("A003", "tt0112453", -0.8, "54654675")],
            columns=["user_id", "item_id", "score", "timestamp"])
        self.ratings = Ratings.from_dataframe(ratings)

        # tt0112281 is rated for A000 but let's suppose we want to know its rank
        self.filter_list = ['tt0112281', 'tt0112760', 'tt0112896']

        movies_dir = os.path.join(dir_test_files, 'complex_contents', 'movies_codified/')

        self.available_loaded_items = LoadedContentsDict(movies_dir)

        # IMPORTANT! If a new classifier is added, just add it to this list to test it
        self.classifiers_list = [
            SkSVC(), SkKNN(), SkRandomForest(), SkLogisticRegression(),
            SkDecisionTree(), SkGaussianProcess()
        ]

    def test_predict(self):
        # Doesn't matter which classifier we chose
        alg = ClassifierRecommender({'Plot': ['tfidf']}, SkSVC(), threshold=0)
        user_ratings = self.ratings.get_user_interactions("A000")
        alg.process_rated(user_ratings, self.available_loaded_items)
        alg.fit()

        # Will raise Exception since it's not a Score Prediction Algorithm
        with self.assertRaises(NotPredictionAlg):
            alg.predict(user_ratings, self.available_loaded_items)

    @for_each_classifier
    def test_rank_single_representation(self, classifier: Classifier):
        clf = classifier

        # Single representation
        alg = ClassifierRecommender({'Plot': ['tfidf']}, clf, threshold=0)

        user_ratings = self.ratings.get_user_interactions("A000")

        alg.process_rated(user_ratings, self.available_loaded_items)
        alg.fit()

        # rank with filter_list
        res_filtered = alg.rank(user_ratings, self.available_loaded_items, filter_list=self.filter_list)
        item_ranked_set = set([interaction_filtered.item_id for interaction_filtered in res_filtered])
        self.assertEqual(len(item_ranked_set), len(self.filter_list))
        self.assertCountEqual(item_ranked_set, self.filter_list)

        # rank without filter_list
        res_all_unrated = alg.rank(user_ratings, self.available_loaded_items)
        item_rated_set = set([interaction.item_id for interaction in user_ratings])
        item_ranked_set = set([interaction_filtered.item_id for interaction_filtered in res_all_unrated])
        # We expect this to be empty, since the alg should rank only unrated items (unless in filter list)
        rated_in_ranked = item_ranked_set.intersection(item_rated_set)
        self.assertEqual(len(rated_in_ranked), 0)

        # rank with n_recs specified
        n_recs = 5
        res_n_recs = alg.rank(user_ratings, self.available_loaded_items, n_recs)
        self.assertEqual(len(res_n_recs), n_recs)
        item_rated_set = set([interaction.item_id for interaction in user_ratings])
        item_ranked_set = set([interaction_nrecs.item_id for interaction_nrecs in res_n_recs])
        # We expect this to be empty, since the alg should rank only unrated items (unless in filter list)
        rated_in_ranked = item_ranked_set.intersection(item_rated_set)
        self.assertEqual(len(rated_in_ranked), 0)

    @for_each_classifier
    def test_rank_multiple_representations(self, classifier: Classifier):
        clf = classifier

        # Multiple representations with auto threshold based on the mean ratings of the user
        alg = ClassifierRecommender({'Plot': ['tfidf', 'embedding'],
                                     'Genre': ['tfidf', 'embedding'],
                                     'imdbRating': [0]}, clf)

        user_ratings = self.ratings.get_user_interactions("A000")

        alg.process_rated(user_ratings, self.available_loaded_items)
        alg.fit()

        # rank with filter_list
        res_filtered = alg.rank(user_ratings, self.available_loaded_items, filter_list=self.filter_list)
        item_ranked_set = set([interaction_filtered.item_id for interaction_filtered in res_filtered])
        self.assertEqual(len(item_ranked_set), len(self.filter_list))
        self.assertCountEqual(item_ranked_set, self.filter_list)

        # rank without filter_list
        res_all_unrated = alg.rank(user_ratings, self.available_loaded_items)
        item_rated_set = set([interaction.item_id for interaction in user_ratings])
        item_ranked_set = set([interaction_filtered.item_id for interaction_filtered in res_all_unrated])
        # We expect this to be empty, since the alg should rank only unrated items (unless in filter list)
        rated_in_ranked = item_ranked_set.intersection(item_rated_set)
        self.assertEqual(len(rated_in_ranked), 0)

        # rank with n_recs specified
        n_recs = 5
        res_n_recs = alg.rank(user_ratings, self.available_loaded_items, n_recs)
        self.assertEqual(len(res_n_recs), n_recs)
        item_rated_set = set([interaction.item_id for interaction in user_ratings])
        item_ranked_set = set([interaction_nrecs.item_id for interaction_nrecs in res_n_recs])
        # We expect this to be empty, since the alg should rank only unrated items (unless in filter list)
        rated_in_ranked = item_ranked_set.intersection(item_rated_set)
        self.assertEqual(len(rated_in_ranked), 0)

    def test_raise_errors(self):
        # Only positive available
        ratings = pd.DataFrame.from_records([
            ("A000", "tt0112281", 1, "54654675")],
            columns=["user_id", "item_id", "score", "timestamp"])
        ratings = Ratings.from_dataframe(ratings)

        alg = ClassifierRecommender({'Plot': 'tfidf'}, SkKNN(), 0)
        user_ratings = ratings.get_user_interactions("A000")

        with self.assertRaises(OnlyPositiveItems):
            alg.process_rated(user_ratings, self.available_loaded_items)

        # Only negative available
        ratings = pd.DataFrame.from_records([
            ("A000", "tt0112281", -1, "54654675")],
            columns=["user_id", "item_id", "score", "timestamp"])
        ratings = Ratings.from_dataframe(ratings)

        alg = ClassifierRecommender({'Plot': 'tfidf'}, SkKNN(), 0)
        user_ratings = ratings.get_user_interactions("A000")

        with self.assertRaises(OnlyNegativeItems):
            alg.process_rated(user_ratings, self.available_loaded_items)

        # No Item avilable locally
        ratings = pd.DataFrame.from_records([
            ("A000", "non existent", 0.5, "54654675")],
            columns=["user_id", "item_id", "score", "timestamp"])
        ratings = Ratings.from_dataframe(ratings)

        alg = ClassifierRecommender({'Plot': 'tfidf'}, SkKNN(), 0)
        user_ratings = ratings.get_user_interactions("A000")

        with self.assertRaises(NoRatedItems):
            alg.process_rated(user_ratings, self.available_loaded_items)

        # User has no ratings
        user_ratings = []

        alg = ClassifierRecommender({'Plot': 'tfidf'}, SkKNN(), 0)

        with self.assertRaises(EmptyUserRatings):
            alg.process_rated(user_ratings, self.available_loaded_items)

        with self.assertRaises(EmptyUserRatings):
            alg.rank(user_ratings, self.available_loaded_items)


if __name__ == '__main__':
    unittest.main()



from . import embedding_technique

from .embedding_technique import *
from .tf_idf import WhooshTfIdf, SkLearnTfIdf
from .field_content_production_technique import OriginalData
from .synset_document_frequency import PyWSDSynsetDocumentFrequency



from os.path import isfile, splitext, join
from os import listdir
from abc import abstractmethod, ABC
from typing import Set, Iterable

from clayrs.content_analyzer.memory_interfaces.text_interface import SearchIndex
from clayrs.utils import load_content_instance
from clayrs.utils.const import logger


class LoadedContentsInterface(ABC):
    @abstractmethod
    def get_contents_interface(self):
        raise NotImplementedError


class LoadedContentsDict(LoadedContentsInterface):

    def __init__(self, contents_path: str, contents_to_load: Set[str] = None, only_representations: dict = None):
        self._contents_path = contents_path

        self._available_items_set = {splitext(filename)[0]
                                     for filename in listdir(contents_path)
                                     if isfile(join(contents_path, filename)) and splitext(filename)[1] == ".xz"}

        # we load all available items
        if contents_to_load is None:
            contents_to_load_present = self._available_items_set
        else:
            contents_to_load = set(contents_to_load)
            contents_to_load_present = self._available_items_set.intersection(set(contents_to_load))
            if len(contents_to_load_present) != len(contents_to_load):
                logger.warning("Some items are not present locally, they can't be loaded")

        self._contents_dict = {}
        if len(contents_to_load_present) != 0:
            logger.info("Loading contents from disk...")
            self._contents_dict = {item_id: load_content_instance(contents_path, item_id, only_representations)
                                   for item_id in contents_to_load_present}

            if not any(self._contents_dict.values()):
                raise FileNotFoundError(f"No contents found in {contents_path}! "
                                        f"Maybe you have misspelled the path folder?")

    def get_contents_interface(self):
        return self._contents_dict

    def get(self, key: str, only_representations: dict = None):
        content = self._contents_dict.get(key)
        if content is None:
            content = load_content_instance(self._contents_path, key, only_representations)
            self._contents_dict[key] = content

        return content

    def get_list(self, key_list: Iterable[str], only_representations: dict = None):
        contents_to_load = set(key_list) - set(self._contents_dict.keys())
        self._contents_dict.update({content: load_content_instance(self._contents_path, content, only_representations)
                                    for content in contents_to_load})

        return [self._contents_dict[content_id] for content_id in key_list]

    def __getitem__(self, key: str):
        return self._contents_dict[key]

    def __iter__(self):
        yield from self._available_items_set

    def __len__(self):
        return len(self._contents_dict)

    def __str__(self):
        return str(self._contents_dict)

    def __repr__(self):
        return repr(self._contents_dict)


class LoadedContentsIndex(LoadedContentsInterface):
    def __init__(self, index_path: str):
        self._contents_index = SearchIndex(index_path)

    def get_contents_interface(self):
        return self._contents_index



from __future__ import annotations
from abc import abstractmethod
from typing import List, Set, Callable, TYPE_CHECKING

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    from clayrs.content_analyzer.ratings_manager.ratings import Interaction
    from clayrs.recsys.partitioning import Split

from clayrs.evaluation.metrics.metrics import Metric


class ClassificationMetric(Metric):
    """
    Abstract class that generalize classification metrics.
    A classification metric uses confusion matrix terminology (true positive, false positive, etc.) to classify each
    item predicted

    Args:
        relevant_threshold (float): parameter needed to discern relevant items and non-relevant items for every
            user. If not specified, the mean rating score of every user will be used
        sys_average (str): specify how the system average must be computed. Default is 'macro'
    """

    def __init__(self, relevant_threshold: float = None, sys_average: str = 'macro',
                 precision: [Callable] = np.float64):
        valid_avg = {'macro', 'micro'}
        self.__avg = sys_average.lower()

        if self.__avg not in valid_avg:
            raise ValueError("Average {} is not supported! Average methods available for {} are:\n"
                             "{}".format(sys_average, str(self), valid_avg))

        self.__relevant_threshold = relevant_threshold
        self.__precision = precision

    @property
    def relevant_threshold(self):
        return self.__relevant_threshold

    @property
    def sys_avg(self):
        return self.__avg

    @property
    def precision(self):
        return self.__precision

    def perform(self, split: Split) -> pd.DataFrame:
        # This method will calculate for every split true positive, false positive, true negative, false negative items
        # so that every metric must simply implement the method _calc_metric(...).
        # Thanks to polymorphism, everything will work without changing this main method

        pred = split.pred
        truth = split.truth

        split_result = {'user_id': [], str(self): []}
        sys_confusion_matrix = np.array([[0, 0],
                                         [0, 0]], dtype=np.int32)

        for user in set(truth.user_id_column):
            user_predictions = pred.get_user_interactions(user)
            user_truth = truth.get_user_interactions(user)

            relevant_threshold = self.relevant_threshold
            if relevant_threshold is None:
                relevant_threshold = np.nanmean([truth_interaction.score
                                                 for truth_interaction in user_truth])

            user_truth_relevant_items = set([truth_interaction.item_id for truth_interaction in user_truth
                                             if truth_interaction.score >= relevant_threshold])

            # If basically the user has not provided a rating to any items greater than the threshold,
            # then we don't consider it since it's not fault of the system
            if len(user_truth_relevant_items) != 0:
                metric_user, user_confusion_matrix = self._perform_single_user(user_predictions,
                                                                               user_truth_relevant_items)

                sys_confusion_matrix += user_confusion_matrix
            else:
                metric_user = np.nan

            split_result['user_id'].append(user)
            split_result[str(self)].append(metric_user)

        # trick to check for nan values, if all values are nan then an exception is thrown
        if all(user_result != user_result for user_result in split_result[str(self)]):
            raise ValueError("No user has a rating above the given threshold! Try lower it")

        sys_metric = -1
        if self.sys_avg == 'micro':
            sys_metric = self._calc_metric(sys_confusion_matrix)
        elif self.sys_avg == 'macro':
            sys_metric = np.nanmean(split_result[str(self)])

        split_result['user_id'].append('sys')
        split_result[str(self)].append(sys_metric)

        return pd.DataFrame(split_result)

    @abstractmethod
    def _calc_metric(self, confusion_matrix: np.ndarray):
        """
        Private method that must be implemented by every metric which must specify how to use the confusion matrix
        terminology in order to compute the metric
        """
        raise NotImplementedError

    @abstractmethod
    def _perform_single_user(self, user_prediction_items: List[Interaction], user_truth_items: Set[str]):
        raise NotImplementedError


class Precision(ClassificationMetric):
    r"""
    The Precision metric is calculated as such for the **single user**:

    $$
    Precision_u = \frac{tp_u}{tp_u + fp_u}
    $$

    Where:

    - $tp_u$ is the number of items which are in the recommendation list of the user and have a
      rating >= relevant_threshold in its 'ground truth'
    - $fp_u$ is the number of items which are in the recommendation list of the user and have a
      rating < relevant_threshold in its 'ground truth'

    And it is calculated as such for the **entire system**, depending if 'macro' average or 'micro' average has been
    chosen:

    $$ 
    Precision_{sys} - micro = \frac{\sum_{u \in U} tp_u}{\sum_{u \in U} tp_u + \sum_{u \in U} fp_u}
    $$
    
    $$
    Precision_{sys} - macro = \frac{\sum_{u \in U} Precision_u}{|U|}
    $$

    Args:
        relevant_threshold: parameter needed to discern relevant items and non-relevant items for every
            user. If not specified, the mean rating score of every user will be used
        sys_average: specify how the system average must be computed. Default is 'macro'
    """
    def __init__(self, relevant_threshold: float = None, sys_average: str = 'macro',
                 precision: [Callable] = np.float64):
        super(Precision, self).__init__(relevant_threshold, sys_average, precision)

    def __str__(self):
        return "Precision - {}".format(self.sys_avg)

    def __repr__(self):
        return f"Precision(relevant_threshold={self.relevant_threshold}, sys_average={self.sys_avg}, " \
               f"precision={self.precision})"

    def _calc_metric(self, confusion_matrix: np.ndarray):
        tp = confusion_matrix[0, 0]
        fp = confusion_matrix[0, 1]
        return self.precision((tp + fp) and tp / (tp + fp) or 0)  # safediv between tp and (tp + fp)

    def _perform_single_user(self, user_prediction: List[Interaction], user_truth_relevant_items: Set[str]):
        tp = len([prediction_interaction for prediction_interaction in user_prediction
                  if prediction_interaction.item_id in user_truth_relevant_items])
        fp = len(user_prediction) - tp

        # we do not compute the full confusion matrix for the user
        useful_confusion_matrix_user = np.array([[tp, fp],
                                                 [0, 0]], dtype=np.int32)

        return self._calc_metric(useful_confusion_matrix_user), useful_confusion_matrix_user


class PrecisionAtK(Precision):
    r"""
    The Precision@K metric is calculated as such for the **single user**:

    $$
    Precision@K_u = \frac{tp@K_u}{tp@K_u + fp@K_u}
    $$

    Where:

    - $tp@K_u$ is the number of items which are in the recommendation list  of the user
      **cutoff to the first K items** and have a rating >= relevant_threshold in its 'ground truth'
    - $tp@K_u$ is the number of items which are in the recommendation list  of the user
      **cutoff to the first K items** and have a rating < relevant_threshold in its 'ground truth'

    And it is calculated as such for the **entire system**, depending if 'macro' average or 'micro' average has been
    chosen:

    $$
    Precision@K_{sys} - micro = \frac{\sum_{u \in U} tp@K_u}{\sum_{u \in U} tp@K_u + \sum_{u \in U} fp@K_u}
    $$

    $$
    Precision@K_{sys} - macro = \frac{\sum_{u \in U} Precision@K_u}{|U|}
    $$

    Args:
        k (int): cutoff parameter. Only the first k items of the recommendation list will be considered
        relevant_threshold (float): parameter needed to discern relevant items and non-relevant items for every
            user. If not specified, the mean rating score of every user will be used
        sys_average (str): specify how the system average must be computed. Default is 'macro'
    """

    def __init__(self, k: int, relevant_threshold: float = None, sys_average: str = 'macro',
                 precision: [Callable] = np.float64):
        super().__init__(relevant_threshold, sys_average, precision)
        if k < 1:
            raise ValueError('k={} not valid! k must be >= 1!'.format(k))
        self.__k = k

    @property
    def k(self):
        return self.__k

    def __str__(self):
        return "Precision@{} - {}".format(self.k, self.sys_avg)

    def __repr__(self):
        return f"PrecisionAtK(k={self.k}, relevant_threshold={self.relevant_threshold}, sys_average={self.sys_avg}, " \
               f"precision={self.precision})"

    def _perform_single_user(self, user_prediction: List[Interaction], user_truth_relevant_items: Set[str]):
        user_prediction_cut = user_prediction[:self.k]

        tp = len([prediction_interaction for prediction_interaction in user_prediction_cut
                  if prediction_interaction.item_id in user_truth_relevant_items])
        fp = len(user_prediction_cut) - tp

        useful_confusion_matrix_user = np.array([[tp, fp],
                                                 [0, 0]], dtype=np.int32)

        return self._calc_metric(useful_confusion_matrix_user), useful_confusion_matrix_user


class RPrecision(Precision):
    r"""
    The R-Precision metric is calculated as such for the **single user**:

    $$
    R-Precision_u = \frac{tp@R_u}{tp@R_u + fp@R_u}
    $$

    Where:

    - $R$ it's the number of relevant items for the user *u*
    - $tp@R_u$ is the number of items which are in the recommendation list  of the user
      **cutoff to the first R items** and have a rating >= relevant_threshold in its 'ground truth'
    - $tp@R_u$ is the number of items which are in the recommendation list  of the user
      **cutoff to the first R items** and have a rating < relevant_threshold in its 'ground truth'

    And it is calculated as such for the **entire system**, depending if 'macro' average or 'micro' average has been
    chosen:

    $$
    Precision@R_{sys} - micro = \frac{\sum_{u \in U} tp@R_u}{\sum_{u \in U} tp@R_u + \sum_{u \in U} fp@R_u}
    $$

    $$
    Precision@R_{sys} - macro = \frac{\sum_{u \in U} R-Precision_u}{|U|}
    $$

    Args:
        relevant_threshold: parameter needed to discern relevant items and non-relevant items for every
            user. If not specified, the mean rating score of every user will be used
        sys_average: specify how the system average must be computed. Default is 'macro'
    """
    def __init__(self, relevant_threshold: float = None, sys_average: str = 'macro',
                 precision: [Callable] = np.float64):
        super().__init__(relevant_threshold, sys_average, precision)

    def __str__(self):
        return "R-Precision - {}".format(self.sys_avg)

    def __repr__(self):
        return f"RPrecision(relevant_threshold={self.relevant_threshold}, sys_average={self.sys_avg}, " \
               f"precision={self.precision})"

    def _perform_single_user(self, user_prediction: List[Interaction], user_truth_relevant_items: Set[str]):
        r = len(user_truth_relevant_items)
        user_prediction_cut = user_prediction[:r]

        tp = len([prediction_interaction for prediction_interaction in user_prediction_cut
                  if prediction_interaction.item_id in user_truth_relevant_items])
        fp = len(user_prediction_cut) - tp

        useful_confusion_matrix_user = np.array([[tp, fp],
                                                 [0, 0]], dtype=np.int32)

        return self._calc_metric(useful_confusion_matrix_user), useful_confusion_matrix_user


class Recall(ClassificationMetric):
    r"""
    The Recall metric is calculated as such for the **single user**:

    $$
    Recall_u = \frac{tp_u}{tp_u + fn_u}
    $$

    Where:

    - $tp_u$ is the number of items which are in the recommendation list of the user and have a
      rating >= relevant_threshold in its 'ground truth'
    - $fn_u$ is the number of items which are NOT in the recommendation list of the user and have a
      rating >= relevant_threshold in its 'ground truth'

    And it is calculated as such for the **entire system**, depending if 'macro' average or 'micro' average has been
    chosen:

    $$
    Recall_{sys} - micro = \frac{\sum_{u \in U} tp_u}{\sum_{u \in U} tp_u + \sum_{u \in U} fn_u}
    $$

    $$
    Recall_{sys} - macro = \frac{\sum_{u \in U} Recall_u}{|U|}
    $$

    Args:
        relevant_threshold (float): parameter needed to discern relevant items and non-relevant items for every
            user. If not specified, the mean rating score of every user will be used
        sys_average (str): specify how the system average must be computed. Default is 'macro'
    """

    def __init__(self, relevant_threshold: float = None, sys_average: str = 'macro',
                 precision: [Callable] = np.float64):
        super().__init__(relevant_threshold, sys_average, precision)

    def __str__(self):
        return "Recall - {}".format(self.sys_avg)

    def __repr__(self):
        return f"Recall(relevant_threshold={self.relevant_threshold}, sys_average={self.sys_avg}, " \
               f"precision={self.precision})"

    def _calc_metric(self, confusion_matrix: np.ndarray):
        tp = confusion_matrix[0, 0]
        fn = confusion_matrix[1, 0]
        return self.precision((tp + fn) and tp / (tp + fn) or 0)  # safediv between tp and (tp + fn)

    def _perform_single_user(self, user_prediction: List[Interaction], user_truth_relevant_items: Set[str]):
        tp = len([prediction_interaction for prediction_interaction in user_prediction
                  if prediction_interaction.item_id in user_truth_relevant_items])
        fn = len(user_truth_relevant_items) - tp

        useful_confusion_matrix_user = np.array([[tp, 0],
                                                 [fn, 0]], dtype=np.int32)

        return self._calc_metric(useful_confusion_matrix_user), useful_confusion_matrix_user


class RecallAtK(Recall):
    r"""
    The Recall@K metric is calculated as such for the **single user**:

    $$
    Recall@K_u = \frac{tp@K_u}{tp@K_u + fn@K_u}
    $$

    Where:

    - $tp@K_u$ is the number of items which are in the recommendation list  of the user
      **cutoff to the first K items** and have a rating >= relevant_threshold in its 'ground truth'
    - $tp@K_u$ is the number of items which are NOT in the recommendation list  of the user
      **cutoff to the first K items** and have a rating >= relevant_threshold in its 'ground truth'

    And it is calculated as such for the **entire system**, depending if 'macro' average or 'micro' average has been
    chosen:

    $$
    Recall@K_{sys} - micro = \frac{\sum_{u \in U} tp@K_u}{\sum_{u \in U} tp@K_u + \sum_{u \in U} fn@K_u}
    $$

    $$
    Recall@K_{sys} - macro = \frac{\sum_{u \in U} Recall@K_u}{|U|}
    $$

    Args:
        k (int): cutoff parameter. Only the first k items of the recommendation list will be considered
        relevant_threshold (float): parameter needed to discern relevant items and non-relevant items for every
            user. If not specified, the mean rating score of every user will be used
        sys_average (str): specify how the system average must be computed. Default is 'macro'
    """

    def __init__(self, k: int, relevant_threshold: float = None, sys_average: str = 'macro',
                 precision: [Callable] = np.float64):
        super().__init__(relevant_threshold, sys_average, precision)
        if k < 1:
            raise ValueError('k={} not valid! k must be >= 1!'.format(k))
        self.__k = k

    @property
    def k(self):
        return self.__k

    def __str__(self):
        return "Recall@{} - {}".format(self.k, self.sys_avg)

    def __repr__(self):
        return f"RecallAtK(k={self.k}, relevant_threshold={self.relevant_threshold}, sys_average={self.sys_avg}, " \
               f"precision={self.precision})"

    def _perform_single_user(self, user_prediction: List[Interaction], user_truth_relevant_items: Set[str]):
        user_prediction_cut = user_prediction[:self.k]

        tp = len([prediction_interaction for prediction_interaction in user_prediction_cut
                  if prediction_interaction.item_id in user_truth_relevant_items])
        fn = len(user_truth_relevant_items) - tp

        useful_confusion_matrix_user = np.array([[tp, 0],
                                                 [fn, 0]], dtype=np.int32)

        return self._calc_metric(useful_confusion_matrix_user), useful_confusion_matrix_user


class FMeasure(ClassificationMetric):
    r"""
    The FMeasure metric combines Precision and Recall into a single metric. It is calculated as such for the
    **single user**:

    $$
    FMeasure_u = (1 + \beta^2) \cdot \frac{P_u \cdot R_u}{(\beta^2 \cdot P_u) + R_u}
    $$

    Where:

    - $P_u$ is the Precision calculated for the user *u*
    - $R_u$ is the Recall calculated for the user *u*
    - $\beta$ is a real factor which could weight differently Recall or Precision based on its value:

        - $\beta = 1$: Equally weight Precision and Recall
        - $\beta > 1$: Weight Recall more
        - $\beta < 1$: Weight Precision more

    A famous FMeasure is the F1 Metric, where $\beta = 1$, which basically is the harmonic mean of recall and
    precision:

    $$
    F1_u = \frac{2 \cdot P_u \cdot R_u}{P_u + R_u}
    $$

    The FMeasure metric is calculated as such for the **entire system**, depending if 'macro' average or 'micro'
    average has been chosen:

    $$
    FMeasure_{sys} - micro = (1 + \beta^2) \cdot \frac{P_u \cdot R_u}{(\beta^2 \cdot P_u) + R_u}
    $$

    $$
    FMeasure_{sys} - macro = \frac{\sum_{u \in U} FMeasure_u}{|U|}
    $$

    Args:
        beta (float): real factor which could weight differently Recall or Precision based on its value. Default is 1
        relevant_threshold (float): parameter needed to discern relevant items and non-relevant items for every
            user. If not specified, the mean rating score of every user will be used
        sys_average (str): specify how the system average must be computed. Default is 'macro'
    """

    def __init__(self, beta: float = 1, relevant_threshold: float = None, sys_average: str = 'macro',
                 precision: [Callable] = np.float64):
        super().__init__(relevant_threshold, sys_average, precision)
        self.__beta = beta

    @property
    def beta(self):
        return self.__beta

    def __str__(self):
        return "F{} - {}".format(self.beta, self.sys_avg)

    def __repr__(self):
        return f"FMeasure(beta={self.beta}, relevant_threshold={self.relevant_threshold}, " \
               f"sys_average={self.sys_avg}, precision={self.precision})"

    def _calc_metric(self, confusion_matrix: np.ndarray):
        prec = Precision()._calc_metric(confusion_matrix)
        reca = Recall()._calc_metric(confusion_matrix)

        beta_2 = self.beta ** 2

        num = prec * reca
        den = (beta_2 * prec) + reca

        fbeta = self.precision((1 + beta_2) * (den and num / den or 0))  # safediv between num and den

        return fbeta

    def _perform_single_user(self, user_prediction: List[Interaction], user_truth_relevant_items: Set[str]):
        tp = len([prediction_interaction for prediction_interaction in user_prediction
                  if prediction_interaction.item_id in user_truth_relevant_items])
        fp = len(user_prediction) - tp
        fn = len(user_truth_relevant_items) - tp

        useful_confusion_matrix_user = np.array([[tp, fp],
                                                 [fn, 0]], dtype=np.int32)

        return self._calc_metric(useful_confusion_matrix_user), useful_confusion_matrix_user


class FMeasureAtK(FMeasure):
    r"""
    The FMeasure@K metric combines Precision@K and Recall@K into a single metric. It is calculated as such for the
    **single user**:

    $$
    FMeasure@K_u = (1 + \beta^2) \cdot \frac{P@K_u \cdot R@K_u}{(\beta^2 \cdot P@K_u) + R@K_u}
    $$

    Where:

    - $P@K_u$ is the Precision at K calculated for the user *u*
    - $R@K_u$ is the Recall at K calculated for the user *u*
    - $\beta$ is a real factor which could weight differently Recall or Precision based on its value:

        - $\beta = 1$: Equally weight Precision and Recall
        - $\beta > 1$: Weight Recall more
        - $\beta < 1$: Weight Precision more

    A famous FMeasure@K is the F1@K Metric, where :math:`\beta = 1`, which basically is the harmonic mean of recall and
    precision:

    $$
    F1@K_u = \frac{2 \cdot P@K_u \cdot R@K_u}{P@K_u + R@K_u}
    $$

    The FMeasure@K metric is calculated as such for the **entire system**, depending if 'macro' average or 'micro'
    average has been chosen:

    $$
    FMeasure@K_{sys} - micro = (1 + \beta^2) \cdot \frac{P@K_u \cdot R@K_u}{(\beta^2 \cdot P@K_u) + R@K_u}
    $$

    $$
    FMeasure@K_{sys} - macro = \frac{\sum_{u \in U} FMeasure@K_u}{|U|}
    $$

    Args:
        k (int): cutoff parameter. Will be used for the computation of Precision@K and Recall@K
        beta (float): real factor which could weight differently Recall or Precision based on its value. Default is 1
        relevant_threshold (float): parameter needed to discern relevant items and non-relevant items for every
            user. If not specified, the mean rating score of every user will be used
        sys_average (str): specify how the system average must be computed. Default is 'macro'
    """

    def __init__(self, k: int, beta: int = 1, relevant_threshold: float = None, sys_average: str = 'macro'):
        super().__init__(beta, relevant_threshold, sys_average)
        if k < 1:
            raise ValueError('k={} not valid! k must be >= 1!'.format(k))
        self.__k = k

    @property
    def k(self):
        return self.__k

    def __str__(self):
        return "F{}@{} - {}".format(self.beta, self.k, self.sys_avg)

    def __repr__(self):
        return f"FMeasureAtK(k={self.k}, beta={self.beta}, relevant_threshold={self.relevant_threshold}, " \
               f"sys_average={self.sys_avg}, precision={self.precision})"

    def _perform_single_user(self, user_prediction: List[Interaction], user_truth_relevant_items: Set[str]):
        user_prediction_cut = user_prediction[:self.k]

        tp = len([prediction_interaction for prediction_interaction in user_prediction_cut
                  if prediction_interaction.item_id in user_truth_relevant_items])
        fp = len(user_prediction_cut) - tp
        fn = len(user_truth_relevant_items) - tp

        useful_confusion_matrix_user = np.array([[tp, fp],
                                                 [fn, 0]], dtype=np.int32)

        return self._calc_metric(useful_confusion_matrix_user), useful_confusion_matrix_user



from unittest import TestCase

import os

import numpy as np
from gensim.corpora import Dictionary
from gensim.models import LdaModel
from gensim.test.utils import common_texts
import gensim

from clayrs.content_analyzer.embeddings.embedding_learner.lda import GensimLDA

# we fix random_state for reproducibility
random_state = 42
num_topics = 100
model_path = 'test_model_lda'


class TestLda(TestCase):
    def test_all(self):
        my_learner = GensimLDA(model_path, num_topics=num_topics, random_state=random_state)

        corpus = common_texts
        my_learner.fit_model(corpus)

        # check that vector size is correct
        self.assertEqual(num_topics, my_learner.get_vector_size())

        common_dictionary = Dictionary(common_texts)
        common_corpus = [common_dictionary.doc2bow(text) for text in common_texts]
        expected_learner = LdaModel(common_corpus, num_topics=num_topics, random_state=random_state)

        # test get_embedding not existent document
        unseen_doc_text = ['this', 'is', 'a', 'new', 'document', 'which', 'doesnt', 'exist']
        unseen_doc = common_dictionary.doc2bow(unseen_doc_text)
        expected = expected_learner[unseen_doc]
        expected_vector: np.ndarray = gensim.matutils.sparse2full(expected, num_topics)

        result_vector = my_learner.get_embedding(unseen_doc_text)

        self.assertTrue(np.array_equal(expected_vector, result_vector))

        # test get_embedding existent document
        unseen_doc_text = ['human', 'time', 'trees']
        unseen_doc = common_dictionary.doc2bow(unseen_doc_text)
        expected = expected_learner[unseen_doc]
        expected_vector: np.ndarray = gensim.matutils.sparse2full(expected, num_topics)

        result_vector = my_learner.get_embedding(unseen_doc_text)

        self.assertTrue(np.array_equal(expected_vector, result_vector))

        # test save
        my_learner.save()
        self.assertTrue(os.path.isfile(f"{model_path}.model"))
        self.assertTrue(os.path.isfile(f"{model_path}.model.expElogbeta.npy"))
        self.assertTrue(os.path.isfile(f"{model_path}.model.id2word"))
        self.assertTrue(os.path.isfile(f"{model_path}.model.state"))

        # test that after load we obtain a valid embedding
        my_learner_loaded = GensimLDA(model_path)
        my_learner_loaded.load_model()
        unseen_doc_text = ['human', 'time', 'trees']
        result_vector = my_learner.get_embedding(unseen_doc_text)

        self.assertTrue(np.any(result_vector))



import os
from typing import Union, Iterator

import distex
import contextlib

from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from clayrs.utils.const import logger


@contextlib.contextmanager
def get_progbar(iterator, total=None) -> tqdm:
    bar_format = "{desc} {percentage:.0f}%|{bar}| {n:}/{total_fmt} [{elapsed}<{remaining}]"
    with logging_redirect_tqdm(loggers=[logger]):
        with tqdm(iterator, bar_format=bar_format, total=total) as pbar:
            yield pbar


def handle_exception(loop, context):
    # this is a simple hack to stopping asyncio from logging "task was never retrieved" exception
    # that should not happen in the first place.
    # In fact this problem happens only on specific scenarios like Pycharm interpreter, or by running
    # an asyncio snippet as script, but does not happen if the exact same script is run interactively,
    # or in IPython environment
    pass


@contextlib.contextmanager
def get_iterator_parallel(num_cpus, f_to_parallelize, *args_to_f,
                          progress_bar=False, total=None) -> Union[Iterator, tqdm]:

    num_cpus = num_cpus or os.cpu_count() or 1

    if num_cpus > 1:
        pool = distex.Pool(num_workers=num_cpus, func_pickle=distex.PickleType.cloudpickle)
        pool._loop.set_exception_handler(handle_exception)
        iterator_res = pool.map(f_to_parallelize, *args_to_f)
    else:
        pool = None
        iterator_res = map(f_to_parallelize, *args_to_f)

    try:
        if progress_bar:
            with get_progbar(iterator_res, total=total) as pbar:
                yield pbar
        else:
            yield iterator_res
    finally:
        if pool is not None:
            pool.shutdown()



from unittest import TestCase

from clayrs.content_analyzer.utils.id_merger import id_values_merger


class Test(TestCase):
    def test_id_merger(self):
        self.assertEqual(id_values_merger('aaa'), 'aaa', "Must return a string value")
        self.assertEqual(id_values_merger(['aaa', 'bbb']), 'aaa_bbb', "Must return a string value like this aaa_bbb")
        self.assertEqual(id_values_merger(123), '123', "Must return a string value")
        self.assertEqual(id_values_merger([123, 124]), '123_124', "Must return a string value like this 123_124")
        self.assertEqual(id_values_merger([123, "aaa"]), '123_aaa', "Must return a string value like 123_aaa")
        with self.assertRaises(TypeError):
            id_values_merger({1: 1, 2: 2})



from unittest import TestCase
from clayrs.content_analyzer.ratings_manager.score_processor import NumberNormalizer


class TestNumberNormalizer(TestCase):
    def test_fit(self):
        scores = [1, 2, 5, 5, 3, 3.5, 3.6, 3.7, 3.8, 3.9, 4.0, 10]

        result = []
        for score in scores:
            converted = NumberNormalizer(scale=(1, 10)).fit(score)
            result.append(converted)

        expected = [-1.0, -0.77777777, -0.11111111, -0.11111111,
                    -0.55555555, -0.44444444, -0.42222222, -0.39999999,
                    -0.37777777, -0.35555555, -0.33333333, 1.0]

        for expected_score, result_score in zip(expected, result):
            self.assertAlmostEqual(expected_score, result_score)

        # Test with rounding at the fourth digit
        result_rounded = []
        for score in scores:
            converted_rounded = NumberNormalizer(scale=(1, 10), decimal_rounding=4).fit(score)
            result_rounded.append(converted_rounded)

        expected_rounded = [-1.0, -0.7778, -0.1111, -0.1111, -0.5556,
                            -0.4444, -0.4222, -0.4, -0.3778, -0.3556,
                            -0.3333, 1.0]

        for expected_score_rounded, result_score_rounded in zip(expected_rounded, result_rounded):
            self.assertAlmostEqual(expected_score_rounded, result_score_rounded)

    def test_error(self):

        # 2 numbers must be passed
        with self.assertRaises(ValueError):
            NumberNormalizer(scale=(1,))

        with self.assertRaises(ValueError):
            NumberNormalizer(scale=(1, 2, 3))



from unittest import TestCase
import numpy as np

from clayrs.content_analyzer.content_representation.representation_container import RepresentationContainer


class TestRepresentationContainer(TestCase):

    def test_rep_container(self):
        rep_container = RepresentationContainer(['rep1', 'rep2', 'rep3'], ['test1', None, 'test3'])

        # tests to check that the indexes and columns of the dataframe in representation container are set as expected
        self.assertEqual([0, 1, 2], rep_container.get_internal_index())
        self.assertEqual(['test1', None, 'test3'], rep_container.get_external_index())
        self.assertEqual(['rep1', 'rep2', 'rep3'], rep_container.get_representations())

        # tests to check that the representation related to the internal_id or external_id passed to rep_container
        # is the appropriate representation
        for _ in range(15000):
            self.assertEqual('rep1', rep_container[0])
            self.assertEqual('rep3', rep_container['test3'])

        # tests to check the correct functionality of the append and remove method
        rep_container.append('rep4', 'test4')
        self.assertEqual('rep4', rep_container['test4'])

        value_removed = rep_container.pop('test4')
        self.assertEqual('rep4', value_removed)
        self.assertFalse('rep4' in rep_container.get_representations())

        # test for empty representation container
        empty_rep_container = RepresentationContainer()
        self.assertEqual(0, len(empty_rep_container))

        # test for passing single value to representation container constructor instead of lists
        single_rep_container = RepresentationContainer('rep', 'test')
        self.assertEqual('rep', single_rep_container['test'])

        # test exception different length of external_id representation lists when passed to the constructor
        with self.assertRaises(ValueError):
            RepresentationContainer(['rep1', 'rep2'], ['test1'])

        # test exception different length of external_id representation lists when passed to the 'append' method
        with self.assertRaises(ValueError):
            rep_container.append(['rep1', 'rep2'], ['test1'])

        # test exception representation not present
        with self.assertRaises(KeyError):
            err = rep_container['not_existent']

    def test_iter(self):
        rep_container = RepresentationContainer(['rep1', 'rep2', 'rep3'], ['test1', None, 'test3'])

        expected_list = [
            {'internal_id': 0, 'external_id': 'test1', 'representation': 'rep1'},
            {'internal_id': 1, 'external_id': None, 'representation': 'rep2'},
            {'internal_id': 2, 'external_id': 'test3', 'representation': 'rep3'}
        ]

        it = iter(rep_container)

        self.assertEqual(expected_list[0], next(it))
        self.assertEqual(expected_list[1], next(it))
        self.assertEqual(expected_list[2], next(it))

        # Check that the iterator gives an error since there aren't any items left
        with self.assertRaises(StopIteration):
            next(it)



from typing import List

import numpy as np
import gensim
from gensim.models import RpModel
from gensim.corpora import Dictionary

from clayrs.content_analyzer.embeddings.embedding_learner.embedding_learner import GensimDocumentEmbeddingLearner
from clayrs.content_analyzer.utils.check_tokenization import check_tokenized


class GensimRandomIndexing(GensimDocumentEmbeddingLearner):
    """
    Class that implements RandomIndexing model thanks to the Gensim library.

    If a pre-trained local Word2Vec model must be loaded, put its path in the `reference` parameter.
    Otherwise, a Word2Vec model will be trained from scratch based on the preprocessed corpus of the contents to complexly
    represent

    If you'd like to save the model once trained, set the path in the `reference` parameter and set
    `auto_save=True`. If `reference` is None, trained model won't be saved after training and will only be used to
    produce contents in the current run

    Additional parameters regarding the model itself could be passed, check [gensim documentation](https://radimrehurek.com/gensim/models/rpmodel.html)
    to see what else can be customized

    Args:
        reference: Path of the model to load/where the model trained will be saved if `auto_save=True`. If None the
            trained model won't be saved after training and will only be used to produce contents in the current run
        auto_save: If True, the model will be saved in the path specified in `reference` parameter
    """

    def __init__(self, reference: str = None, auto_save: bool = True, **kwargs):
        super().__init__(reference, auto_save, ".model", **kwargs)

    def fit_model(self, corpus: List):
        dictionary = Dictionary(corpus)
        word_docs_matrix = [dictionary.doc2bow(doc) for doc in corpus]
        self.model = RpModel(word_docs_matrix, id2word=dictionary, **self.additional_parameters)

    def load_model(self):
        return RpModel.load(self.reference)

    def get_vector_size(self) -> int:
        return self.model.num_topics

    def get_embedding(self, document_tokenized: List[str]) -> np.ndarray:
        unseen_doc = self.model.id2word.doc2bow(check_tokenized(document_tokenized))

        # if document is totally new (no word in train corpus) KeyError is raised
        # and load method of embedding source will fill the document vector with zeros
        if len(unseen_doc) == 0:
            raise KeyError

        sparse_vector = self.model[unseen_doc]
        dense_vector = gensim.matutils.sparse2full(sparse_vector, self.model.num_topics)
        return dense_vector

    def __str__(self):
        return "GensimRandomProjections"

    def __repr__(self):
        return f"GensimRandomProjections(reference={self.reference}, auto_save={self._auto_save}, " \
               f"{', '.join(f'{arg}={val}' for arg, val in self._additional_parameters.items())})"



from unittest import TestCase

from clayrs.content_analyzer.config import ContentAnalyzerConfig, FieldConfig, ItemAnalyzerConfig, \
    UserAnalyzerConfig
from clayrs.utils.class_utils import get_all_implemented_classes, get_all_implemented_subclasses


class TestClassUtils(TestCase):

    def test_get_all_implemented_classes(self):

        results = get_all_implemented_classes(ContentAnalyzerConfig)

        expected_results = {ItemAnalyzerConfig, UserAnalyzerConfig}
        self.assertEqual(results, expected_results)

        results = get_all_implemented_classes(FieldConfig)

        expected_results = {FieldConfig}
        self.assertEqual(results, expected_results)

    def test_get_all_implemented_subclasses(self):

        results = get_all_implemented_subclasses(ContentAnalyzerConfig)

        expected_results = {ItemAnalyzerConfig, UserAnalyzerConfig}
        self.assertEqual(results, expected_results)

        results = get_all_implemented_subclasses(FieldConfig)
        expected_results = set()

        self.assertEqual(results, expected_results)


import abc
from abc import ABC


class Algorithm(ABC):
    """
    Abstract class for an Algorithm.

    Every algorithm must be able to predict or to rank, or maybe both.
    In case some algorithms can only do one of the two (eg. PageRank), simply implement both
    methods and raise the NotPredictionAlg or NotRankingAlg exception accordingly.
    """
    __slots__ = ()

    @abc.abstractmethod
    def predict(self, **kwargs):
        """
        Method to call when score prediction must be done.

        If the Algorithm can't do score prediction, implement this method and raise
        the NotPredictionAlg exception
        """
        raise NotImplementedError

    @abc.abstractmethod
    def rank(self, **kwargs):
        """
        Method to call when ranking must be done.

        If the Algorithm can't rank, implement this method and raise the NotRankingAlg exception
        """
        raise NotImplementedError



import unittest
import pandas as pd
import numpy as np

from clayrs.content_analyzer import Ratings, Rank
from clayrs.evaluation.exceptions import NotEnoughUsers
from clayrs.evaluation.metrics.fairness_metrics import PredictionCoverage, CatalogCoverage, GiniIndex, \
    DeltaGap, GroupFairnessMetric, Counter
from clayrs.evaluation.eval_pipeline_modules.metric_evaluator import Split

# Will be the same for every test
truth = pd.DataFrame({'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u2', 'u2', 'u2'],
                      'item_id': ['i1', 'i2', 'i3', 'i4', 'i6', 'i1', 'i8', 'i4'],
                      'score': [3, 2, 3, 1, 2, 4, 2, 3]})
truth = Ratings.from_dataframe(truth)


class TestPredictionCoverage(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        catalog = {'i1', 'i2', 'i3', 'i4', 'i5', 'i6', 'i7', 'i8', 'i9', 'i10'}
        cls.metric = PredictionCoverage(catalog)

    def test_perform_only_new(self):
        metric = self.metric

        pred_only_new_items = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u2', 'u2'],
             'item_id': ['inew1', 'inew2', 'inew3', 'inew4'],
             'score': [650, 600, 500, 650]})
        pred_only_new_items = Ratings.from_dataframe(pred_only_new_items)

        split_only_new = Split(pred_only_new_items, truth)

        expected = 0
        result = float(metric.perform(split_only_new)[str(metric)])

        self.assertEqual(expected, result)

    def test_perform_w_new_items(self):
        metric = self.metric

        pred_w_new_items = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u2', 'u2', 'u2', 'u2'],
             'item_id': ['i2', 'i1', 'i4', 'i5', 'i6', 'i3', 'i8', 'i9', 'i4', 'i6', 'i1', 'i8'],
             'score': [650, 600, 500, 400, 300, 220, 100, 50, 350, 200, 100, 50]})
        pred_w_new_items = Ratings.from_dataframe(pred_w_new_items)

        split_w_new_items = Split(pred_w_new_items, truth)

        result = float(metric.perform(split_w_new_items)[str(metric)])

        self.assertTrue(0 <= result <= 100)

    def test_perform_all_items(self):
        metric = self.metric

        pred_all_items = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u2', 'u2', 'u2'],
             'item_id': ['i2', 'i1', 'i4', 'i5', 'i6', 'i3', 'i8', 'i9', 'i4', 'i10', 'i7'],
             'score': [650, 600, 500, 400, 300, 220, 100, 50, 350, 200, 100]})
        pred_all_items = Ratings.from_dataframe(pred_all_items)

        split_all = Split(pred_all_items, truth)

        expected = 100
        result = float(metric.perform(split_all)[str(metric)])

        self.assertEqual(expected, result)

    def test_perform_all_items_and_plus(self):
        catalog = {'i1'}
        metric = PredictionCoverage(catalog)

        # All catalog plus more
        pred_all_items_and_plus = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u1'],
             'item_id': ['i2', 'i1', 'i4', 'i5'],
             'score': [650, 600, 500, 400, ]})
        pred_all_items_and_plus = Ratings.from_dataframe(pred_all_items_and_plus)

        split_all = Split(pred_all_items_and_plus, truth)

        expected = 100
        result = float(metric.perform(split_all)[str(metric)])

        self.assertEqual(expected, result)


class TestCatalogCoverage(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        catalog = {'i1', 'i2', 'i3', 'i4', 'i5', 'i6', 'i7', 'i8', 'i9', 'i10'}
        cls.metric_top_n = CatalogCoverage(catalog, top_n=2)
        cls.metric_k_sampling = CatalogCoverage(catalog, k=1)

    def test_perform_only_new(self):
        metric_top_n = self.metric_top_n
        metric_k_sampling = self.metric_k_sampling

        pred_only_new_items = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u2', 'u2'],
             'item_id': ['inew1', 'inew2', 'inew3', 'inew4'],
             'score': [650, 600, 500, 650]})
        pred_only_new_items = Ratings.from_dataframe(pred_only_new_items)

        split_only_new = Split(pred_only_new_items, truth)

        expected = 0

        result_top_n = float(metric_top_n.perform(split_only_new)[str(metric_top_n)])
        self.assertEqual(expected, result_top_n)

        result_k_sampling = float(metric_k_sampling.perform(split_only_new)[str(metric_k_sampling)])
        self.assertEqual(expected, result_k_sampling)

    def test_perform_w_new_items(self):
        metric_top_n = self.metric_top_n
        metric_k_sampling = self.metric_k_sampling

        pred_w_new_items = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u2', 'u2', 'u2', 'u2'],
             'item_id': ['i2', 'i1', 'i4', 'i5', 'i6', 'i3', 'i8', 'i9', 'i4', 'i6', 'i1', 'i8'],
             'score': [650, 600, 500, 400, 300, 220, 100, 50, 350, 200, 100, 50]})
        pred_w_new_items = Ratings.from_dataframe(pred_w_new_items)

        split_w_new_items = Split(pred_w_new_items, truth)

        result_top_n = float(metric_top_n.perform(split_w_new_items)[str(metric_top_n)])
        self.assertTrue(0 <= result_top_n <= 100)

        result_k_sampling = float(metric_k_sampling.perform(split_w_new_items)[str(metric_k_sampling)])
        self.assertTrue(0 <= result_k_sampling <= 100)

    def test_perform_all_items(self):
        catalog = {'i2'}

        metric_top_n = CatalogCoverage(catalog, top_n=1)
        metric_k_sampling = CatalogCoverage(catalog, k=1, top_n=1)

        pred_all_items = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u2', 'u2'],
             'item_id': ['i2', 'i1', 'i2', 'i7'],
             'score': [650, 600, 200, 100]})
        pred_all_items = Ratings.from_dataframe(pred_all_items)

        split_all = Split(pred_all_items, truth)

        expected = 100

        result = float(metric_top_n.perform(split_all)[str(metric_top_n)])
        self.assertEqual(expected, result)

        result = float(metric_k_sampling.perform(split_all)[str(metric_k_sampling)])
        self.assertEqual(expected, result)

    def test_perform_all_items_and_plus(self):
        catalog = {'i2'}

        metric_top_n = CatalogCoverage(catalog, top_n=2)
        metric_k_sampling = CatalogCoverage(catalog, k=1)

        pred_all_items_and_plus = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u2', 'u2'],
             'item_id': ['i2', 'i1', 'i2', 'i7'],
             'score': [650, 600, 200, 100]})
        pred_all_items_and_plus = Ratings.from_dataframe(pred_all_items_and_plus)

        split_all = Split(pred_all_items_and_plus, truth)

        expected = 100

        result = float(metric_top_n.perform(split_all)[str(metric_top_n)])
        self.assertEqual(expected, result)

        result = float(metric_k_sampling.perform(split_all)[str(metric_k_sampling)])
        self.assertEqual(expected, result)


class TestGini(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.metric = GiniIndex()

        cls.metric_top_n = GiniIndex(top_n=2)

    def test_perform_equi(self):
        metric = self.metric

        # i1 and i2 and i3 are recommended in equal ways to users
        pred_equi = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u2', 'u2', 'u2'],
             'item_id': ['i1', 'i2', 'i3', 'i2', 'i3', 'i1'],
             'score': [650, 600, 500, 750, 700, 680]})
        pred_equi = Ratings.from_dataframe(pred_equi)

        split_pred_equi = Split(pred_equi, truth)

        expected = 0
        result = float(metric.perform(split_pred_equi)[str(metric)])

        self.assertEqual(expected, result)

    def test_perform_equi_top_n(self):
        metric_top_n = self.metric_top_n

        # i1 and i2 and i3 are recommended in equal ways to users
        pred_equi = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u2', 'u2', 'u2'],
             'item_id': ['i1', 'i2', 'i3', 'i2', 'i3', 'i1'],
             'score': [650, 600, 500, 750, 700, 680]})
        pred_equi = Ratings.from_dataframe(pred_equi)

        split_pred_equi = Split(pred_equi, truth)

        result = float(metric_top_n.perform(split_pred_equi)[str(metric_top_n)])

        # In the top 2 i2 is recommended more, so there's no equality
        self.assertTrue(0 < result <= 1)

        metric_top_3 = GiniIndex(top_n=3)

        result = float(metric_top_3.perform(split_pred_equi)[str(metric_top_3)])

        # In the top 3 (total length of rec lists) there's equality
        self.assertEqual(0, result)

    def test_perform_mixed(self):
        metric = self.metric
        metric_top_n = self.metric_top_n

        pred_mixed = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u2', 'u2', 'u2', 'u2'],
             'item_id': ['i2', 'i1', 'i4', 'i5', 'i6', 'i3', 'i8', 'i9', 'i2', 'i6', 'i1', 'i8'],
             'score': [650, 600, 500, 400, 300, 220, 100, 50, 350, 200, 100, 50]})
        pred_mixed = Ratings.from_dataframe(pred_mixed)

        split_mixed = Split(pred_mixed, truth)

        result = float(metric.perform(split_mixed)[str(metric)])
        self.assertTrue(0 < result <= 1)

        result = float(metric_top_n.perform(split_mixed)[str(metric_top_n)])
        self.assertTrue(0 < result <= 1)

    def test_perform_only_one(self):
        metric = self.metric
        metric_top_n = self.metric_top_n

        pred_only_one_item = pd.DataFrame(
            {'user_id': ['u1'],
             'item_id': ['i4'],
             'score': [650]})
        pred_only_one_item = Ratings.from_dataframe(pred_only_one_item)

        split_only_one = Split(pred_only_one_item, truth)

        expected = 0
        result = float(metric.perform(split_only_one)[str(metric)])

        self.assertEqual(expected, result)

        # Even if there's only one element and top_n = 2, everything works correctly
        expected = 0
        result = float(metric_top_n.perform(split_only_one)[str(metric_top_n)])

        self.assertEqual(expected, result)


class TestGroupFairnessMetric(unittest.TestCase):
    def test_split_users_in_groups(self):
        pred_only_u1 = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1'],
             'item_id': ['i2', 'i1', 'i4'],
             'score': [650, 600, 500]})
        popular_items = {'i2'}
        pred_only_u1 = Ratings.from_dataframe(pred_only_u1)

        # Check error raised when split 1 users in 2 groups (n_users < n_groups)
        with self.assertRaises(NotEnoughUsers):
            GroupFairnessMetric.split_user_in_groups(pred_only_u1,
                                                     groups={'a': 0.5, 'b': 0.5},
                                                     pop_items=popular_items)

        # Check error raised a percentage is not valid
        with self.assertRaises(ValueError):
            GroupFairnessMetric.split_user_in_groups(pred_only_u1,
                                                     groups={'a': 1.9},
                                                     pop_items=popular_items)

        # Check error raised when percentage total < 1
        pred_4_users = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u2', 'u2', 'u2', 'u3', 'u3', 'u4', 'u4', 'u4', 'u4'],
             'item_id': ['i2', 'i1', 'i4', 'i5', 'i6', 'i3', 'i8', 'i9', 'i2', 'i6', 'i1', 'i8'],
             'score': [650, 600, 500, 400, 300, 220, 100, 50, 350, 200, 100, 50]})
        popular_items = {'i2', 'i1'}
        pred_4_users = Ratings.from_dataframe(pred_4_users)

        with self.assertRaises(ValueError):
            result = GroupFairnessMetric.split_user_in_groups(pred_4_users,
                                                              groups={'a': 0.3, 'b': 0.5},
                                                              pop_items=popular_items)

        # Check splitted groups in a usual situation
        result = GroupFairnessMetric.split_user_in_groups(pred_4_users,
                                                          groups={'a': 0.5, 'b': 0.3, 'c': 0.2},
                                                          pop_items=popular_items)

        # u1 and u4 are the users that prefer popular items,
        # so they are put into the first group
        self.assertIn('u1', result['a'])
        self.assertIn('u4', result['a'])

        # u2 and u3 have the same popolarity ratio, so we are not sure if they are put
        # into the 2nd group or the third one
        self.assertTrue('u2' in result['b'] or 'u2' in result['c'])
        self.assertTrue('u3' in result['b'] or 'u3' in result['c'])

    def test_get_avg_pop_by_users(self):
        pred_4_users = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u2', 'u2', 'u2', 'u3', 'u3', 'u4', 'u4', 'u4', 'u4'],
             'item_id': ['i2', 'i1', 'i4', 'i5', 'i6', 'i3', 'i8', 'i9', 'i2', 'i6', 'i1', 'i8'],
             'score': [650, 600, 500, 400, 300, 220, 100, 50, 350, 200, 100, 50]})

        pop_by_item = Counter(list(pred_4_users['item_id']))

        # sum of popularity of every item rated by the user / number of rated items by the user
        # (the popularity of an item is the number of times it is repeated in the frame)
        expected_u1 = 5 / 3
        expected_u2 = 4 / 3
        expected_u3 = 3 / 2
        expected_u4 = 8 / 4

        pred_4_users = Ratings.from_dataframe(pred_4_users)
        # Calculate avg_pop for every user of the frame
        result_all = GroupFairnessMetric.get_avg_pop_by_users(pred_4_users, pop_by_item)

        # Check that the resuts are corrected
        self.assertAlmostEqual(expected_u1, result_all['u1'])
        self.assertAlmostEqual(expected_u2, result_all['u2'])
        self.assertAlmostEqual(expected_u3, result_all['u3'])
        self.assertAlmostEqual(expected_u4, result_all['u4'])

        # Calculate avg_pop for only users in the 'group' parameter passed
        result_u1_u2 = GroupFairnessMetric.get_avg_pop_by_users(pred_4_users, pop_by_item, group={'u1', 'u2'})

        # Check that the results are correct
        self.assertAlmostEqual(expected_u1, result_all['u1'])
        self.assertAlmostEqual(expected_u2, result_all['u2'])
        # Check that 'u3' and 'u4' are not present
        self.assertNotIn('u3', result_u1_u2.keys())
        self.assertNotIn('u4', result_u1_u2.keys())


class TestDeltaGap(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        original_ratings = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1',
                         'u2', 'u2', 'u2', 'u2',
                         'u3', 'u3', 'u3', 'u3', 'u3', 'u3',
                         'u4', 'u4', 'u4', 'u4', 'u4', 'u4', 'u4',
                         'u5', 'u5', 'u5', 'u5', 'u5',
                         'u6', 'u6'],
             'item_id': ['i1', 'i2', 'i3', 'i4', 'i5', 'i6', 'i7', 'i8',
                         'i1', 'i9', 'i10', 'i11',
                         'i1', 'i12', 'i13', 'i3', 'i10', 'i14',
                         'i3', 'i10', 'i15', 'i16', 'i9', 'i17', 'i99',
                         'i10', 'i18', 'i19', 'i20', 'i21',
                         'inew_1', 'inew_2'],
             'score': [5, 4, 4, 1, 2, 3, 3, 1,
                       4, 5, 1, 1,
                       3, 3, 2, 1, 1, 4,
                       4, 4, 5, 5, 3, 3, 3,
                       3, 3, 2, 2, 1,
                       4, 3]})
        cls.original_ratings = Ratings.from_dataframe(original_ratings)

        train = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1',  # removed last 2
                         'u2', 'u2', 'u2',  # removed last 1
                         'u3', 'u3', 'u3', 'u3',  # removed last 2
                         'u4', 'u4', 'u4', 'u4', 'u4',  # removed last 2
                         'u5', 'u5', 'u5', 'u5',  # removed last 1
                         'u6'],  # removed last 1
             'item_id': ['i1', 'i2', 'i3', 'i4', 'i5', 'i6',
                         'i1', 'i9', 'i10',
                         'i1', 'i12', 'i13', 'i3',
                         'i3', 'i10', 'i15', 'i16', 'i9',
                         'i10', 'i18', 'i19', 'i20',
                         'inew_1'],
             'score': [5, 4, 4, 1, 2, 3,
                       4, 5, 1,
                       3, 3, 2, 1,
                       4, 4, 5, 5, 3,
                       3, 3, 2, 2,
                       4]})
        cls.train = Ratings.from_dataframe(train)

        truth = pd.DataFrame(
            {'user_id': ['u1', 'u1',
                         'u2',
                         'u3', 'u3',
                         'u4', 'u4', 'u4',
                         'u5',
                         'u6'],
             'item_id': ['i7', 'i8',
                         'i11',
                         'i10', 'i14',
                         'i9', 'i17', 'i99',
                         'i21',
                         'inew_2'],
             'score': [3, 1,
                       1,
                       1, 4,
                       3, 3, 3,
                       1,
                       3]})
        truth = Ratings.from_dataframe(truth)

        # u6 is missing, just to test DeltaGap in case for some users recs can't be computed
        recs = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1',
                         'u2', 'u2', 'u2', 'u2', 'u2',
                         'u3', 'u3', 'u3', 'u3', 'u3',
                         'u4', 'u4', 'u4', 'u4', 'u4',
                         'u5', 'u5', 'u5', 'u5', 'u5'],
             'item_id': ['i7', 'i10', 'i11', 'i12', 'i13',
                         'i11', 'i20', 'i6', 'i3', 'i4',
                         'i4', 'i5', 'i6', 'i7', 'i10',
                         'i9', 'i2', 'i3', 'i1', 'i5',
                         'i2', 'i3', 'i4', 'i5', 'i6'],
             'score': [500, 400, 300, 200, 100,
                       400, 300, 200, 100, 50,
                       150, 125, 110, 100, 80,
                       390, 380, 360, 320, 200,
                       250, 150, 190, 100, 50]})
        recs = Rank.from_dataframe(recs)

        cls.split = Split(recs, truth)

    def test_calculate_gap(self):
        # This is basically the inner part of the GAP equation, the fraction at the numerator
        # of the GAP formula calculated preemptively for every user.
        # It is set manually for the sake of the test, but it is obtained by the
        # get_avg_pop_by_users() method of the GroupFairnessMetric Class
        avg_pop_by_users = {'u1': 2, 'u2': 1.78, 'u3': 3.5, 'u4': 1.1}

        expected_u1_u3 = (avg_pop_by_users['u1'] + avg_pop_by_users['u3']) / 2
        result_u1_u3 = DeltaGap.calculate_gap({'u1', 'u3'}, avg_pop_by_users)

        self.assertAlmostEqual(expected_u1_u3, result_u1_u3)

        expected_u2_u4 = (avg_pop_by_users['u2'] + avg_pop_by_users['u4']) / 2
        result_u2_u4 = DeltaGap.calculate_gap({'u2', 'u4'}, avg_pop_by_users)

        self.assertAlmostEqual(expected_u2_u4, result_u2_u4)

    def test_calculate_delta_gap(self):
        gap_profile = 2.32
        gap_recs = 3

        result = DeltaGap.calculate_delta_gap(gap_recs, gap_profile)
        expected = (gap_recs - gap_profile) / gap_profile

        self.assertAlmostEqual(expected, result)

    def test_invalid_percentage(self):
        with self.assertRaises(ValueError):
            DeltaGap(user_groups={'a': 0.5}, pop_percentage=-0.5, user_profiles=self.train,
                     original_ratings=self.original_ratings)
            DeltaGap(user_groups={'a': 0.5}, pop_percentage=0, user_profiles=self.train,
                     original_ratings=self.original_ratings)
            DeltaGap(user_groups={'a': 0.5}, pop_percentage=1.5, user_profiles=self.train,
                     original_ratings=self.original_ratings)

    def _compute_deltagap(self, valid_groups_splitted: dict, top_n: int = None):

        n_users = len(set(self.original_ratings.user_id_column))  # { u1, u2, u3, u4, u5, u6 } so 6 users
        pop_by_item = {item_id: count / n_users
                       for item_id, count in Counter(self.original_ratings.item_id_column).items()}

        expected_result_list = []

        for group_name, valid_group in valid_groups_splitted.items():
            # *******************************************************************************************
            # *** For every user in the group calculate the average popularity of the recommendations ***
            # *******************************************************************************************

            # compute for each user of the group the popularity sum in the recommendations
            # (cut the ranking list if top_n != None)
            RECS_sum_pop_group = {
                user: sum([pop_by_item.get(interaction.item_id)
                           for interaction in self.split.pred.get_user_interactions(user)][:top_n])

                for user in valid_group
            }

            # compute for each user of the group the average popularity in the recommendations
            # (sum_pop_item_recommended / n_item_recommended)
            # (cut the ranking list if top_n != None)
            RECS_avg_pop_group = {user: sum_pop / len(self.split.pred.get_user_interactions(user)[:top_n])
                                  for user, sum_pop in RECS_sum_pop_group.items()}

            # ************************************************************************************
            # *** For every user in the group calculate the average popularity of the profiles ***
            # ************************************************************************************

            # compute for each user of the group the popularity sum in the recommendations
            PROFILE_sum_pop_group = {
                user: sum([pop_by_item.get(interaction.item_id)
                           for interaction in self.train.get_user_interactions(user)])

                for user in valid_group
            }

            # compute for each user of the group the average popularity in the recommendations
            # (sum_pop_item_recommended / n_item_recommended)
            PROFILE_avg_pop_group = {user: sum_pop / len(self.train.get_user_interactions(user))
                                     for user, sum_pop in PROFILE_sum_pop_group.items()}

            # ************************
            # *** Compute DeltaGAP ***
            # ************************

            # sum the RECS_avg_pop of every user of the group / n_users in group
            RECS_gap_group = sum(RECS_avg_pop_group.values()) / len(valid_group)

            # sum the PROFILE_avg_pop of every user of the group / n_users in group
            PROFILE_gap_group = sum(PROFILE_avg_pop_group.values()) / len(valid_group)

            expected_delta_gap_group = (RECS_gap_group - PROFILE_gap_group) / PROFILE_gap_group

            expected_result_list.append(expected_delta_gap_group)

        return expected_result_list

    def test_perform_1_group(self):
        metric = DeltaGap(user_groups={'a': 1},
                          user_profiles=self.train,
                          original_ratings=self.original_ratings)
        result = metric.perform(self.split)

        # since one group, all users will belong to this group
        # except u6 for which we don't have any recs
        valid_group_a = {'u1', 'u2', 'u3', 'u4', 'u4', 'u5'}

        valid_groups_splitted = {'a': valid_group_a}

        [expected_delta_gap_group_a] = self._compute_deltagap(valid_groups_splitted)

        result_delta_gap_group_a = float(result["{} | a".format(str(metric))])

        self.assertAlmostEqual(expected_delta_gap_group_a, result_delta_gap_group_a)

    def test_perform_2_groups(self):
        metric = DeltaGap(user_groups={'a': 0.4, 'b': 0.6},
                          user_profiles=self.train,
                          original_ratings=self.original_ratings)
        result = metric.perform(self.split)

        # since u2 and u4 have higher popular ratio, it is put into the first group)
        valid_group_a = {'u2', 'u4'}

        # u6 won't be considered in computation since we don't have recs for it,
        # but it should belong to group b
        valid_group_b = {'u3', 'u1', 'u5'}

        valid_groups_splitted = {'a': valid_group_a, 'b': valid_group_b}

        [expected_delta_gap_group_a, expected_delta_gap_group_b] = self._compute_deltagap(valid_groups_splitted)

        result_delta_gap_group_a = float(result["{} | a".format(str(metric))])
        result_delta_gap_group_b = float(result["{} | b".format(str(metric))])

        self.assertAlmostEqual(expected_delta_gap_group_a, result_delta_gap_group_a)
        self.assertAlmostEqual(expected_delta_gap_group_b, result_delta_gap_group_b)

    def test_perform_3_group_without_recs(self):
        metric = DeltaGap(user_groups={'a': 0.2, 'b': 0.6, 'c': 0.2},
                          user_profiles=self.train,
                          original_ratings=self.original_ratings)
        result = metric.perform(self.split)

        # no 'c' group since u6 belongs to this group but we don't have any recs for it,
        # so it won't be considered and a warning is printed
        valid_groups_splitted = {'a': {'u2'},
                                 'b': {'u4', 'u5', 'u1', 'u3'}}

        [expected_delta_gap_group_a,
         expected_delta_gap_group_b] = self._compute_deltagap(valid_groups_splitted)

        result_delta_gap_group_a = float(result["{} | a".format(str(metric))])
        result_delta_gap_group_b = float(result["{} | b".format(str(metric))])

        self.assertAlmostEqual(expected_delta_gap_group_a, result_delta_gap_group_a)
        self.assertAlmostEqual(expected_delta_gap_group_b, result_delta_gap_group_b)

    def test_perform_0_gap(self):
        # DeltaGap with 2 equals frame should return 0 for every group
        equal_split = Split(self.split.pred, self.split.pred)

        metric = DeltaGap(user_groups={'a': 0.5, 'b': 0.5},
                          user_profiles=self.train,
                          original_ratings=self.original_ratings)

        result = metric.perform(equal_split)

        for col in result.columns:
            self.assertTrue(v == 0 for v in result[col])

    def test_perform_top_3(self):

        top_n = 3

        metric = DeltaGap(user_groups={'a': 0.4, 'b': 0.6},
                          user_profiles=self.train,
                          original_ratings=self.original_ratings,
                          top_n=top_n)
        result = metric.perform(self.split)

        # since u2 and u4 have higher popular ratio, it is put into the first group)
        valid_group_a = {'u2', 'u4'}

        # u6 won't be considered in computation since we don't have recs for it,
        # but it should belong to group b
        valid_group_b = {'u3', 'u1', 'u5'}

        valid_groups_splitted = {'a': valid_group_a, 'b': valid_group_b}

        [expected_delta_gap_group_a, expected_delta_gap_group_b] = self._compute_deltagap(valid_groups_splitted, top_n)

        result_delta_gap_group_a = float(result["{} | a".format(str(metric))])
        result_delta_gap_group_b = float(result["{} | b".format(str(metric))])

        self.assertAlmostEqual(expected_delta_gap_group_a, result_delta_gap_group_a)
        self.assertAlmostEqual(expected_delta_gap_group_b, result_delta_gap_group_b)

    def test_perform_increased_pop_percentage(self):

        result_pop_normal = DeltaGap(user_groups={'a': 0.4, 'b': 0.6},
                                     user_profiles=self.train,
                                     original_ratings=self.original_ratings).perform(self.split)

        result_pop_increased = DeltaGap(user_groups={'a': 0.4, 'b': 0.6},
                                        user_profiles=self.train,
                                        original_ratings=self.original_ratings,
                                        pop_percentage=0.6).perform(self.split)

        result_pop_normal = np.array(result_pop_normal)
        result_pop_increased = np.array(result_pop_increased)

        result_pop_normal.sort(axis=0)
        result_pop_increased.sort(axis=0)

        # Just check that results with pop_percentage increased are different,
        # since users are put into groups differently
        self.assertFalse(np.array_equal(result_pop_normal, result_pop_increased))

    def test_perform_2_groups_2_splits(self):
        metric = DeltaGap(user_groups={'a': 0.4, 'b': 0.6},
                          user_profiles=[self.train, self.train],
                          original_ratings=self.original_ratings)

        # we are simulating two splits
        for _ in range(2):
            result = metric.perform(self.split)

            # since u2 and u4 have higher popular ratio, it is put into the first group)
            valid_group_a = {'u2', 'u4'}

            # u6 won't be considered in computation since we don't have recs for it,
            # but it should belong to group b
            valid_group_b = {'u3', 'u1', 'u5'}

            valid_groups_splitted = {'a': valid_group_a, 'b': valid_group_b}

            [expected_delta_gap_group_a, expected_delta_gap_group_b] = self._compute_deltagap(valid_groups_splitted)

            result_delta_gap_group_a = float(result["{} | a".format(str(metric))])
            result_delta_gap_group_b = float(result["{} | b".format(str(metric))])

            self.assertAlmostEqual(expected_delta_gap_group_a, result_delta_gap_group_a)
            self.assertAlmostEqual(expected_delta_gap_group_b, result_delta_gap_group_b)

    def test_perform_repeated(self):
        metric = DeltaGap(user_groups={'a': 0.4, 'b': 0.6},
                          user_profiles=self.train,
                          original_ratings=self.original_ratings)

        # we are simulating two splits but only one user profile metric given to the metric
        metric.perform(self.split)
        metric.perform(self.split)


if __name__ == '__main__':
    unittest.main()



import unittest

import ekphrasis.dicts.emoticons

from clayrs.content_analyzer.information_processor.ekphrasis import Ekphrasis


class TestEkphrasis(unittest.TestCase):
    def test_omit_normalize(self):
        # we normalize email and percentages but we omit only emails
        ek = Ekphrasis(omit=['email'], normalize=['email', 'percent'])

        expected = ['this', 'is', 'an', 'email', ':', 'and', 'this', 'is', 'a', 'percent', '<percent>']
        result = ek.process("this is an email: alias@mail.com and this is a percent 23%")

        self.assertEqual(expected, result)

    def test_unpack_contractions(self):
        ek = Ekphrasis(unpack_contractions=True)

        expected = ['i', 'can', 'not', 'do', 'this', 'because', 'i', 'will', 'not', 'and', 'i', 'should', 'not']
        result = ek.process("I can't do this because I won't and I shouldn't")

        self.assertEqual(expected, result)

    def test_unpack_hashtag(self):
        # default corpus, is the english one
        ek = Ekphrasis(unpack_hashtags=True)

        expected = ['next', 'gamedev', 'retrogaming', 'cool', 'photo', 'no', 'unpack']
        result = ek.process("#next #gamedev #retrogaming #coolphoto no unpack")

        self.assertEqual(expected, result)

        # twitter corpus
        ek = Ekphrasis(unpack_hashtags=True, segmenter='twitter')

        expected = ['next', 'game', 'dev', 'retro', 'gaming', 'cool', 'photo', 'no', 'unpack']
        result = ek.process("#next #gamedev #retrogaming #coolphoto no unpack")

        self.assertEqual(expected, result)

    def test_annotate_all_caps_tag(self):
        ek = Ekphrasis(annotate=['allcaps', 'repeated'])

        expected = ['this', 'is', 'good', '!', '<repeated>', 'text', 'and', 'a', '<allcaps>', 'shouted', '</allcaps>',
                    'one']
        result = ek.process("this is good !!! text and a SHOUTED one")

        self.assertEqual(expected, result)

        # change mode for 'allcaps' annotation
        ek = Ekphrasis(annotate=['allcaps', 'repeated'], all_caps_tag='single')

        expected = ['this', 'is', 'good', '!', '<repeated>', 'text', 'and', 'a', 'shouted', '<allcaps>', 'one']
        result = ek.process("this is good !!! text and a SHOUTED one")

        self.assertEqual(expected, result)

    def test_segmentation(self):
        # perform segmentation with the default segmenter corpus ('english')
        ek = Ekphrasis(segmentation=True)

        expected = ['the', 'water', 'cooler', 'exponential', 'backoff', 'no', 'segmentation']
        result = ek.process("thewatercooler exponentialbackoff no segmentation")

        self.assertEqual(expected, result)

        # perform segmentation with 'twitter' as segmenter corpus
        ek = Ekphrasis(segmentation=True, segmenter='twitter')

        expected = ['the', 'watercooler', 'exponential', 'back', 'off', 'no', 'segmentation']
        result = ek.process("thewatercooler exponentialbackoff no segmentation")

        self.assertEqual(expected, result)

        # verify that segmentation segments also hashtags (without removing the #)
        ek = Ekphrasis(segmentation=True, segmenter='twitter')

        expected = ['the', 'watercooler', 'exponential', 'back', 'off', '#', 'cool', 'photo', 'no', 'segmentation']
        result = ek.process("thewatercooler exponentialbackoff #coolphoto no segmentation")

        self.assertEqual(expected, result)

    def test_dicts(self):
        # test additional substitutions with emoticons dict of ekphrasis and a custom dict
        ekphrasis_emoticons_dict = ekphrasis.dicts.emoticons.emoticons
        my_substit_dict = {'you': '<my_sub>'}

        ek = Ekphrasis(dicts=[ekphrasis_emoticons_dict, my_substit_dict])

        expected = ["hello", "<happy>", "<happy>", "how", "are", "<my_sub>", "?", "<sad>", "<sad>", "!"]
        result = ek.process("Hello :-) :) how are you? :( :'(!")

        self.assertEqual(expected, result)

    def test_spell_correction(self):
        # perform spell_correction with the default correction corpus ('english')
        ek = Ekphrasis(spell_correction=True)

        expected = ["this", 'is', 'huuuuge', '.', 'the', 'correct', "way", "of", "doing", "things", "is", 'not', 'the',
                    'following']
        result = ek.process("This is huuuuge. The korrect way of doing tihngs is not the followingt")

        self.assertEqual(expected, result)

        # perform spell_correction with the a different corpus ('twitter')
        ek = Ekphrasis(spell_correction=True, corrector='twitter')
        expected = ["this", 'is', 'huuuuge', 'i', 'the', 'korrect', "way", "of", "doing", "things", "is", 'not', 'the',
                    'following']
        result = ek.process("This is huuuuge. The korrect way of doing tihngs is not the followingt")

        self.assertEqual(expected, result)

        # perform spell_correction also on elongated words
        ek = Ekphrasis(spell_correction=True, spell_correct_elong=True)
        expected = ["this", 'is', 'huge', '.', 'the', 'correct', "way", "of", "doing", "things", "is", 'not', 'the',
                    'following']
        result = ek.process("This is huuuuge. The korrect way of doing tihngs is not the followingt")

        self.assertEqual(expected, result)

    def test_tokenizer(self):
        def my_tokenizer(text):
            return ['test']

        ek = Ekphrasis(tokenizer=my_tokenizer)

        expected = ['test']
        result = ek.process('this is a trial')

        self.assertEqual(expected, result)

    def test_process(self):
        # test all operations
        ek = Ekphrasis(omit=['user'], normalize=['email', 'user'], unpack_contractions=True,
                       unpack_hashtags=True, annotate=['repeated', 'allcaps'], segmenter='twitter', segmentation=True,
                       corrector='english', spell_correction=True, spell_correct_elong=True, all_caps_tag='single')

        expected = ['this', 'should', 'work', 'and', 'this', 'is', 'an', 'email', '<email>', '.',
                    'this', 'can', 'not', 'be', 'and', 'this', 'is', 'as', 'cool', 'test', '.', 'game',
                    'dev', 'must', 'be', 'segmented', 'and', 'this', 'must', 'be', 'corrected', ',', 'even',
                    'elongated', 'words', '.', 'caps', '<allcaps>']

        result = ek.process("This should work @user and this is an email ajeje@mail.com. "
                            "This can't be and this is a #cooltest. gamedev must be segmented and this must "
                            "be korrected, even elongaaaated words. CAPS")

        self.assertEqual(expected, result)



from __future__ import annotations
from collections.abc import Iterable
from typing import List, Set, Union, Dict, TYPE_CHECKING

from clayrs.utils.const import logger

if TYPE_CHECKING:
    from clayrs.content_analyzer import Ratings, Content
    from clayrs.recsys.graphs.graph import Node

from clayrs.recsys.content_based_algorithm.contents_loader import LoadedContentsDict
from clayrs.recsys.graphs.nx_implementation.nx_bipartite_graphs import NXBipartiteGraph
from clayrs.recsys.graphs.graph import TripartiteDiGraph, ItemNode, UserNode
from clayrs.recsys.graphs.graph import PropertyNode
from clayrs.utils.context_managers import get_progbar


# Multiple Inheritance so that we will use NXBipartite as its proper father class (we'll its __init__ method)
# and TripartiteDiGraph as its interface
class NXTripartiteGraph(NXBipartiteGraph, TripartiteDiGraph):
    """
    Class that implements a Tripartite graph through networkx library.

    !!! info

        A *Tripartite Graph* is a graph which supports *User* nodes, *Item* nodes and *Property* nodes, but the latter
        can only be linked to *Item* nodes.
        If you need maximum flexibility, consider using a Full Graph

    It creates a graph from an initial Rating object.

    Consider the following matrix representation of the Rating object
    ```
        +------+-----------+-------+
        | User |   Item    | Score |
        +------+-----------+-------+
        | u1   | Tenet     |     4 |
        | u2   | Inception |     5 |
        | ...  | ...       |   ... |
        +------+-----------+-------+
    ```

    The graph will be created with the following interactions:

    ```
                 4
            u1 -----> Tenet
                 5
            u2 -----> Inception
    ```

    where `u1` and `u2` become *User nodes* and `Tenet` and `Inception` become *Item nodes*,
    with the edge weighted depending on the score given

    If the `link_label` parameter is specified, then each link between users and items will be labeled with the label
    specified (e.g. `link_label='score'`):

    ```
            (4, 'score')
        u1 -------------> Tenet
            (5, 'score')
        u2 -------------> Inception
    ```

    Then the framework tries to load 'Tenet' and 'Inception' from the `item_contents_dir` if it is specified and if
    succeeds, adds in the graph their loaded properties as specified in the `item_exo_properties` parameter.

    !!! info "Load exogenous properties"

        In order to load properties in the graph, we must specify where items are serialized and ***which
        properties to add***:

        *   If *item_exo_properties* is specified as a **set**, then the graph will try to load **all properties**
        from **said exogenous representation**

        ```python
        {'my_exo_id'}
        ```

        *   If *item_exo_properties* is specified as a **dict**, then the graph will try to load **said properties**
        from **said exogenous representation**

        ```python
        {'my_exo_id': ['my_prop1', 'my_prop2']]}
        ```

    Args:
        source_frame: The initial Ratings object needed to create the graph
        item_exo_properties: Set or Dict which contains representations to load from items. Use a `Set` if you want
            to load all properties from specific representations, or use a `Dict` if you want to choose which properties
            to load from specific representations
        item_contents_dir: The path containing items serialized with the Content Analyzer
        link_label: If specified, each link will be labeled with the given label. Default is None

    """

    def __init__(self, source_frame: Ratings = None,
                 item_exo_properties: Union[Dict, set] = None,
                 item_contents_dir: str = None,
                 link_label: str = None):

        NXBipartiteGraph.__init__(self, source_frame, link_label)

        if item_exo_properties and not item_contents_dir:
            logger.warning("`item_exo_properties` parameter set but `item_contents_dir` is None! "
                           "No property will be loaded")
        elif not item_exo_properties and item_contents_dir:
            logger.warning("`item_contents_dir` parameter set but `item_exo_properties` is None! "
                           "No property will be loaded")

        if source_frame is not None and item_contents_dir is not None and item_exo_properties is not None:
            self.add_node_with_prop([ItemNode(item_id) for item_id in set(source_frame.item_id_column)],
                                    item_exo_properties,
                                    item_contents_dir)

    @property
    def property_nodes(self) -> Set[PropertyNode]:
        """
        Returns a set of all *Property nodes* in the graph
        """
        return set(node for node in self._graph.nodes if isinstance(node, PropertyNode))

    def add_node(self, node: Union[Node, List[Node]]):
        """
        Adds one or multiple Node objects to the graph.
        Since this is a Tripartite Graph, only `User Node`, `Item Node` and `Property Node` can be added!

        No duplicates are allowed, but different category nodes with same id are (e.g. `ItemNode('1')` and
        `UserNode('1')`)

        Args:
            node: Node(s) object(s) that needs to be added to the graph

        Raises:
            ValueError: Exception raised when one of the node to add to the graph is not a User, Item or Property node
        """
        if not isinstance(node, list):
            node = [node]

        if any(not isinstance(n, (UserNode, ItemNode, PropertyNode)) for n in node):
            raise ValueError("You can only add UserNodes or ItemNodes to a bipartite graph!")

        self._graph.add_nodes_from(node)

    def add_node_with_prop(self, node: Union[ItemNode, List[ItemNode]], item_exo_properties: Union[Dict, set],
                           item_contents_dir: str,
                           item_filename: Union[str, List[str]] = None):
        """
        Adds one or multiple Node objects and its/their properties to the graph.
        Since this is a Tripartite Graph, only `Item Node` are allowed to have properties!

        In order to load properties in the graph, we must specify where items are serialized and ***which
        properties to add***:

        *   If *item_exo_properties* is specified as a **set**, then the graph will try to load **all properties**
        from **said exogenous representation**

        ```python
        {'my_exo_id'}
        ```

        *   If *item_exo_properties* is specified as a **dict**, then the graph will try to load **said properties**
        from **said exogenous representation**

        ```python
        {'my_exo_id': ['my_prop1', 'my_prop2']]}
        ```

        In case you want your node to have a different id from serialized contents, via the `item_filename` parameter
        you can specify what is the filename of the node that you are adding, e.g.

        ```
        item_to_add = ItemNode('different_id')

        # item_filename is 'item_serialized_1.xz'

        graph.add_node_with_prop(item_to_add, ..., item_filename='item_serialized_1')
        ```

        In case you are adding a list of nodes, you can specify the filename for each node in the list.

        Args:
            node: Node(s) object(s) that needs to be added to the graph along with their properties
            item_exo_properties: Set or Dict which contains representations to load from items. Use a `Set` if you want
                to load all properties from specific representations, or use a `Dict` if you want to choose which
                properties to load from specific representations
            item_contents_dir: The path containing items serialized with the Content Analyzer
            item_filename: Filename(s) of the node(s) to add

        Raises:
            ValueError: Exception raised when one of the node to add to the graph with their properties is not
                an ItemNode
        """
        def node_prop_link_generator():
            for n, id in zip(progbar, item_filename):
                item: Content = loaded_items.get(id)

                if item is not None:
                    exo_props = self._get_exo_props(item_exo_properties, item)

                    single_item_prop_edges = [(n,
                                               PropertyNode(prop_dict[prop]),
                                               {'label': prop})
                                              for prop_dict in exo_props for prop in prop_dict]

                else:
                    single_item_prop_edges = []

                yield from single_item_prop_edges

        if not isinstance(node, list):
            node = [node]

        if any(not isinstance(n, ItemNode) for n in node):
            raise ValueError("Only item nodes can be linked to property nodes in a Tripartite Graph!")

        if isinstance(item_exo_properties, set):
            item_exo_properties = dict.fromkeys(item_exo_properties, None)

        if item_filename is None:
            item_filename = [n.value for n in node]

        if not isinstance(item_filename, list):
            item_filename = [item_filename]

        loaded_items = LoadedContentsDict(item_contents_dir, contents_to_load=set(item_filename))
        with get_progbar(node) as progbar:

            progbar.set_description("Creating Item->Properties links")
            self._graph.add_edges_from((tuple_to_add for tuple_to_add in node_prop_link_generator()))

    def _get_exo_props(self, desired_exo_dict: Dict, item: Content):
        extracted_prop = []
        for exo_id, prop_list_desired in desired_exo_dict.items():
            all_exo_prop: dict = item.get_exogenous_representation(exo_id).value
            # here we filter properties if user specified only certain properties to get
            if prop_list_desired is not None:
                filtered_keys = all_exo_prop.keys() & set(prop_list_desired)
                all_exo_prop = {prop: all_exo_prop[prop] for prop in filtered_keys}

            # we must split eventual properties that are lists
            # eg. {film_director: ['tarantino uri', 'nolan uri']} ->
            # [{film_director: 'tarantino uri'}, {film_director: 'nolan uri'}]
            prop_list_val = {prop_label: all_exo_prop[prop_label] for prop_label in all_exo_prop
                             if isinstance(all_exo_prop[prop_label], Iterable) and
                             not isinstance(all_exo_prop[prop_label], str)}
            if len(prop_list_val) != 0:
                splitted_dict = [{prop_label: prop_val} for prop_label in prop_list_val
                                 for prop_val in prop_list_val[prop_label]]

                prop_single_val = {prop_label: all_exo_prop[prop_label]
                                   for prop_label in all_exo_prop.keys() - prop_list_val.keys()}

                extracted_prop.extend(splitted_dict)
                if len(prop_single_val) != 0:
                    extracted_prop.append(prop_single_val)
            else:
                extracted_prop.append(all_exo_prop)

        return extracted_prop

    def add_link(self, start_node: Union[Node, List[Node]], final_node: Union[Node, List[Node]],
                 weight: float = None, label: str = None, timestamp: str = None):
        """
        Creates a link connecting the `start_node` to the `final_node`. If two lists are passed, then the node in
        position $i$ in the `start_node` list will be linked to the node in position $i$ in the `final_node` list.

        If nodes to link do not exist, they will be added automatically to the graph. Please remember that since this is
        a Tripartite Graph, only *User nodes*, *Item nodes* and *Property nodes* can be added! And *Property nodes* can
        only be linked to *Item nodes*!

        A link can be weighted with the `weight` parameter and labeled with the `label` parameter.
        A timestamp can also be specified via `timestamp` parameter.
        All three are optional parameters, so they are not required

        Args:
            start_node: Single Node object or a list of Node objects. They will be the 'head' of the link, since it's a
                directed graph
            final_node (object): Single Node object or a list Node objects. They will be the 'tail' of the link,
                since it's a directed graph
            weight: weight of the link, default is None (no weight)
            label: label of the link, default is None (no label)
            timestamp: timestamp of the link, default is None (no timestamp)

        Raises:
            ValueError: Exception raised when Property nodes are tried to be linked with non-Item nodes
        """

        def is_not_valid_link(start_n: Node, final_n: Node):
            return (isinstance(final_n, PropertyNode) and not isinstance(start_n, ItemNode)) or \
                   (isinstance(start_n, PropertyNode) and not isinstance(final_n, ItemNode))

        if not isinstance(start_node, list):
            start_node = [start_node]

        if not isinstance(final_node, list):
            final_node = [final_node]

        if any(is_not_valid_link(start_n, final_n) for start_n, final_n in zip(start_node, final_node)):
            raise ValueError("Only item nodes can be linked to property nodes in a Tripartite Graph!")

        self.add_node(start_node)
        self.add_node(final_node)

        not_none_dict = {}
        if label is not None:
            not_none_dict['label'] = label
        if weight is not None:
            not_none_dict['weight'] = weight
        if timestamp is not None:
            not_none_dict['timestamp'] = timestamp

        self._graph.add_edges_from(zip(start_node, final_node),
                                   **not_none_dict)

    def __str__(self):
        return "NXTripartiteGraph"

    def __repr__(self):
        return str(self)



from __future__ import annotations
from abc import abstractmethod

import numpy as np
from typing import Union, List, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from clayrs.content_analyzer.content_representation.content import FieldRepresentation
    from clayrs.content_analyzer.embeddings.embedding_learner.embedding_learner import WordEmbeddingLearner, \
        SentenceEmbeddingLearner, DocumentEmbeddingLearner
    from clayrs.content_analyzer.field_content_production_techniques.embedding_technique.combining_technique import \
        CombiningTechnique
    from clayrs.content_analyzer.embeddings.embedding_loader.embedding_loader import \
        WordEmbeddingLoader, SentenceEmbeddingLoader, DocumentEmbeddingLoader
    from clayrs.content_analyzer.embeddings.embedding_loader.embedding_loader import EmbeddingSource
    from clayrs.content_analyzer.information_processor.information_processor import InformationProcessor
    from clayrs.content_analyzer.raw_information_source import RawInformationSource

from clayrs.content_analyzer.embeddings.embedding_learner.embedding_learner import EmbeddingLearner
from clayrs.content_analyzer.content_representation.content import EmbeddingField
from clayrs.content_analyzer.embeddings.embedding_loader.embedding_loader import EmbeddingLoader, EmbeddingSource
from clayrs.content_analyzer.field_content_production_techniques.field_content_production_technique import \
    SingleContentTechnique
from clayrs.content_analyzer.utils.check_tokenization import check_tokenized, tokenize_in_sentences, check_not_tokenized
from clayrs.utils.class_utils import get_all_implemented_subclasses
from clayrs.utils.const import logger
from clayrs.utils.context_managers import get_progbar


class EmbeddingTechnique(SingleContentTechnique):
    """
    Abstract class that generalizes the techniques that create embedding vectors. The EmbeddingTechniques can be split
    into different categories depending on the type of granularity the technique has. For example, a word granularity
    embedding will have a resulting matrix where each row refers to a specific word in the text.

    Args:
        embedding_source (EmbeddingSource): Source where the embeddings vectors for the words in field_data
            are stored.
    """

    def __init__(self, embedding_source: EmbeddingSource):

        super().__init__()

        self.__embedding_source = embedding_source

    @staticmethod
    def from_str_to_embedding_source(embedding_source_str: str, loader_class: Type[EmbeddingLoader]) -> EmbeddingSource:
        """
        Method used to convert a string (which represents a model name) to a corresponding Embedding Source that can
        use the defined model
        Given the loader class (which is a class inheriting from EmbeddingSourceLoader), the method checks each
        implemented class inheriting from loader_class and returns the one that is able to load the model

        The method raises FileNotFoundError if no embedding loader is able to load the model

        Be careful when passing a string, because the first Loader able to load the model will be kept. So if there are
        multiple loaders capable of loading a model, the first one found will be used

        Args:
            embedding_source_str (str): string representing the model name ('twitter-glove-25' for example)
            loader_class (Type): class that inherits from EmbeddingSourceLoader, each subclass of this class will be
                tested with the model name (stored in embedding_source_str)

        Returns:
            embedding_source (EmbeddingSource): embedding source which can load the model
        """
        # retrieves all implementations (meaning not abstract classes) inheriting from loader_class
        possible_implementations = get_all_implemented_subclasses(loader_class)
        # each implementation is tested and, when one of the implementations loads the model successfully, the loader
        # instance is returned
        for implementation in possible_implementations:
            embedding_source = implementation(embedding_source_str)
            if embedding_source.model is not None:
                logger.info("The EmbeddingSource %s was found for the %s reference"
                            % (implementation.__name__, embedding_source_str))
                return embedding_source
        # if no class was found to process the model an exception is raised
        raise FileNotFoundError("The system couldn't process %s as a valid embedding reference"
                                % embedding_source_str)

    @property
    def embedding_source(self):
        return self.__embedding_source

    def produce_content(self, field_name: str, preprocessor_list: List[InformationProcessor],
                        source: RawInformationSource) -> List[FieldRepresentation]:
        representation_list: List[FieldRepresentation] = []

        if isinstance(self.__embedding_source, EmbeddingLoader) and self.__embedding_source.model is None:
            raise FileNotFoundError("The reference %s was not valid for the %s source" %
                                    (self.__embedding_source.reference, self.__embedding_source))

        # if the embedding source is an EmbeddingLearner (meaning it can be trained) and the source has no model
        # the source is trained
        if isinstance(self.__embedding_source, EmbeddingLearner) and self.__embedding_source.model is None:
            logger.warning("The model %s wasn't found, so it will be created and trained now" %
                           self.__embedding_source.reference)
            logger.warning("The model will be trained on the %s field "
                           "and the data will be processed with %s" % (field_name, preprocessor_list))
            self.__embedding_source.fit(source, [field_name], preprocessor_list)

        # it iterates over all contents contained in the source in order to retrieve the raw data
        # the data contained in the field_name is processed using each information processor in the processor_list
        # the data is passed to the method that will create the single representation
        with get_progbar(list(source)) as pbar:

            for content_data in pbar:

                pbar.set_description(f"Processing and producing contents with {self.__embedding_source}")

                processed_data = self.process_data(content_data[field_name], preprocessor_list)
                representation_list.append(self.produce_single_repr(processed_data))

        self.embedding_source.unload_model()
        return representation_list

    @abstractmethod
    def produce_single_repr(self, field_data: Union[List[str], str]) -> EmbeddingField:
        """
        Method that builds the semantic content starting from the embeddings contained in field_data which can
        be either a list of tokens or the entire text in string form (each technique will then have to apply
        processing operations to the data, in order to obtain it in the desired form)

        Args:
            field_data (List[str], str): Data contained in the field, it can be either a list of tokens or the
                original data in string form

        Returns:
            EmbeddingField: complex representation created using the field data
        """
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError


class StandardEmbeddingTechnique(EmbeddingTechnique):
    """
    Class that generalizes the embedding techniques working with the corresponding EmbeddingSource
    StandardEmbeddingTechnique can be extended to consider different types of granularity (word, sentence, ...)
    and each technique should refer to the corresponding EmbeddingSource, so, for example, the
    WordEmbeddingTechnique should refer to the WordEmbeddingSource.

    To sum it up, this class contains the techniques that don't apply any kind of additional operation other
    than loading the embedding from the source
    """

    def __init__(self, embedding_source: EmbeddingSource):
        super().__init__(embedding_source)

    def produce_single_repr(self, field_data: Union[List[str], str]) -> EmbeddingField:
        return EmbeddingField(self.embedding_source.load(self.process_data_granularity(field_data)))

    @abstractmethod
    def process_data_granularity(self, field_data: Union[List[str], str]) -> List[str]:
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError


class WordEmbeddingTechnique(StandardEmbeddingTechnique):
    """
    Class that makes use of a *word* granularity embedding source to produce *word* embeddings

    Args:
        embedding_source: Any `WordEmbedding` model
    """

    def __init__(self, embedding_source: Union[WordEmbeddingLoader, WordEmbeddingLearner, str]):
        # if isinstance(embedding_source, str):
        #     embedding_source = self.from_str_to_embedding_source(embedding_source, WordEmbeddingLoader)
        super().__init__(embedding_source)

    def process_data_granularity(self, field_data: Union[List[str], str]) -> List[str]:
        return check_tokenized(field_data)

    def __str__(self):
        return "WordEmbeddingTechnique"

    def __repr__(self):
        return f'WordEmbeddingTechnique(embedding_source={self.embedding_source})'


class SentenceEmbeddingTechnique(StandardEmbeddingTechnique):
    """
    Class that makes use of a *sentence* granularity embedding source to produce *sentence* embeddings

    Args:
        embedding_source: Any `SentenceEmbedding` model
    """

    def __init__(self, embedding_source: Union[SentenceEmbeddingLoader, SentenceEmbeddingLearner, str]):
        # if isinstance(embedding_source, str):
        #     embedding_source = self.from_str_to_embedding_source(embedding_source, SentenceEmbeddingLoader)
        super().__init__(embedding_source)

    def process_data_granularity(self, field_data: Union[List[str], str]) -> List[str]:
        return tokenize_in_sentences(field_data)

    def __str__(self):
        return "SentenceEmbeddingTechnique"

    def __repr__(self):
        return f'SentenceEmbeddingTechnique(embedding_source={self.embedding_source})'


class DocumentEmbeddingTechnique(StandardEmbeddingTechnique):
    """
    Class that makes use of a *document* granularity embedding source to produce *document* embeddings

    Args:
        embedding_source: Any `DocumentEmbedding` model
    """

    def __init__(self, embedding_source: Union[DocumentEmbeddingLoader, DocumentEmbeddingLearner, str]):
        # if isinstance(embedding_source, str):
        #     embedding_source = self.from_str_to_embedding_source(embedding_source, DocumentEmbeddingLoader)
        super().__init__(embedding_source)

    def process_data_granularity(self, field_data: Union[List[str], str]) -> List[str]:
        return [check_not_tokenized(field_data)]

    def __str__(self):
        return "DocumentEmbeddingTechnique"

    def __repr__(self):
        return f'DocumentEmbeddingTechnique(embedding_source={self.embedding_source})'


class CombiningEmbeddingTechnique(EmbeddingTechnique):
    """
    Class that generalizes the embedding techniques working with a combining technique
    CombiningEmbeddingTechnique can be extended to consider different types of granularity (word, sentence, ...)
    and each technique should be further extended in order to consider the different embedding sources that
    can be combined into the defined granularity.
    For example, with the document granularity, it is possible to combine the embedding matrix of any granularity with a
    lower scope, so word or sentence. With the sentence granularity, it is possible to combine the word embeddings but
    not the document ones (because document's scope is bigger than sentence's scope).

    To sum it up, this class contains the techniques that make use of an EmbeddingSource with different granularity from
    theirs and apply a combining technique to create their embedding matrix

    Args:
        combining_technique (CombiningTechnique): technique that will be used to combine the embeddings retrieved from
        the source
    """

    def __init__(self, embedding_source: EmbeddingSource, combining_technique: CombiningTechnique):
        super().__init__(embedding_source)
        self.__combining_technique = combining_technique

    @property
    def combining_technique(self) -> CombiningTechnique:
        return self.__combining_technique

    @abstractmethod
    def produce_single_repr(self, field_data: Union[List[str], str]) -> EmbeddingField:
        raise NotImplementedError

    @abstractmethod
    def process_data_granularity(self, field_data: Union[List[str], str]) -> List[str]:
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError


class CombiningSentenceEmbeddingTechnique(CombiningEmbeddingTechnique):
    """
    Class that generalizes the combining embedding techniques with sentence granularity
    """

    def __init__(self, embedding_source: EmbeddingSource, combining_technique: CombiningTechnique):
        super().__init__(embedding_source, combining_technique)

    def produce_single_repr(self, field_data: Union[List[str], str]) -> EmbeddingField:
        """
        Produces a single representation with sentence granularity by combining the embedding vectors in order to
        create an embedding matrix that represents the document
        """
        # text is split into sentences
        sentences = tokenize_in_sentences(field_data)

        # the sentences embedding matrix is created and the combined vectors added
        sentences_embeddings = np.ndarray(shape=(len(sentences), self.embedding_source.get_vector_size()))
        for i, sentence in enumerate(sentences):
            sentence_matrix = self.embedding_source.load(self.process_data_granularity(sentence))
            sentences_embeddings[i, :] = self.combining_technique.combine(sentence_matrix)

        return EmbeddingField(sentences_embeddings)

    @abstractmethod
    def process_data_granularity(self, field_data: Union[List[str], str]) -> List[str]:
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError


class Word2SentenceEmbedding(CombiningSentenceEmbeddingTechnique):
    """
    Class that makes use of a word granularity embedding source to produce sentence embeddings

    Args:
        embedding_source: Any `WordEmbedding` model
        combining_technique: Technique used to combine embeddings of finer granularity (word-level) to obtain embeddings
            of coarser granularity (sentence-level)
    """

    def __init__(self, embedding_source: Union[WordEmbeddingLoader, WordEmbeddingLearner, str],
                 combining_technique: CombiningTechnique):
        # if isinstance(embedding_source, str):
        #     embedding_source = self.from_str_to_embedding_source(embedding_source, WordEmbeddingLoader)
        super().__init__(embedding_source, combining_technique)

    def process_data_granularity(self, field_data: Union[List[str], str]) -> List[str]:
        return check_tokenized(field_data)

    def __str__(self):
        return "Word2SentenceEmbedding"

    def __repr__(self):
        return f"Word2SentenceEmbedding(embedding_source={self.embedding_source}, " \
               f"combining_technique={self.combining_technique})"


class CombiningDocumentEmbeddingTechnique(CombiningEmbeddingTechnique):
    """
    Class that generalizes the combining embedding techniques with document granularity
    """

    def __init__(self, embedding_source: EmbeddingSource, combining_technique: CombiningTechnique):
        super().__init__(embedding_source, combining_technique)

    def produce_single_repr(self, field_data: Union[List[str], str]) -> EmbeddingField:
        """
        Produces a single representation with document granularity by combining the embedding vectors in order to
        create an embedding matrix that represents the document
        """
        doc_matrix = self.embedding_source.load(self.process_data_granularity(check_not_tokenized(field_data)))
        return EmbeddingField(self.combining_technique.combine(doc_matrix))

    @abstractmethod
    def process_data_granularity(self, data: Union[List[str], str]) -> List[str]:
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError


class Word2DocEmbedding(CombiningDocumentEmbeddingTechnique):
    """
    Class that makes use of a *word* granularity embedding source to produce embeddings of *document* granularity

    Args:
        embedding_source: Any `WordEmbedding` model
        combining_technique: Technique used to combine embeddings of finer granularity (word-level) to obtain embeddings
            of coarser granularity (doc-level)
    """

    def __init__(self, embedding_source: Union[WordEmbeddingLoader, WordEmbeddingLearner, str],
                 combining_technique: CombiningTechnique):
        # if isinstance(embedding_source, str):
        #     embedding_source = self.from_str_to_embedding_source(embedding_source, WordEmbeddingLoader)
        super().__init__(embedding_source, combining_technique)

    def process_data_granularity(self, field_data: Union[List[str], str]) -> List[str]:
        return check_tokenized(field_data)

    def __str__(self):
        return "Word2DocEmbedding"

    def __repr__(self):
        return f"Word2DocEmbedding(embedding_source={self.embedding_source}, " \
               f"combining_technique={self.combining_technique})"


class Sentence2DocEmbedding(CombiningDocumentEmbeddingTechnique):
    """
    Class that makes use of a *sentence* granularity embedding source to produce embeddings of *document* granularity


    Args:
        embedding_source: Any `SentenceEmbedding` model
        combining_technique: Technique used to combine embeddings of finer granularity (sentence-level) to obtain
            embeddings of coarser granularity (doc-level)
    """

    def __init__(self, embedding_source: Union[SentenceEmbeddingLoader, SentenceEmbeddingLearner, str],
                 combining_technique: CombiningTechnique):
        # if isinstance(embedding_source, str):
        #     embedding_source = self.from_str_to_embedding_source(embedding_source, SentenceEmbeddingLoader)
        super().__init__(embedding_source, combining_technique)

    def process_data_granularity(self, field_data: Union[List[str], str]) -> List[str]:
        return tokenize_in_sentences(field_data)

    def __str__(self):
        return "Sentence2DocEmbedding"

    def __repr__(self):
        return f"Sentence2DocEmbedding(embedding_source={self.embedding_source}, " \
               f"combining_technique={self.combining_technique})"


class DecombiningEmbeddingTechnique(EmbeddingTechnique):
    """
    Class generalizing embedding techniques that contain methods for extracting embedding at finer granularity.
    DecombiningEmbeddingTechnique can be extended to consider Sentence or Document granularity.
    Each technique should be further extended to consider the different sources of embedding that can be combined into
    the defined granularity.
    For example, with Sentence granularity, it is possible, if the model is capable, to extrapolate embedding with word
    granularity.

    To summarize, this class contains techniques that make use of an EmbeddingSource with a granularity different from
    their own, and have the ability to extract the embedding matrix with finer granularity

    """

    def __init__(self, embedding_source: EmbeddingSource):
        super().__init__(embedding_source)

    @abstractmethod
    def produce_single_repr(self, field_data: Union[List[str], str]) -> EmbeddingField:  # return array numpy
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError


class DecombiningInWordsEmbeddingTechnique(DecombiningEmbeddingTechnique):
    """
    Class that generalizes the decombining embedding techniques from a coarse granularity to the finest granularity word
    """

    def __init__(self, embedding_source: Union[SentenceEmbeddingLoader, SentenceEmbeddingLearner,
                                               DocumentEmbeddingLoader, DocumentEmbeddingLearner]):
        super().__init__(embedding_source)

    @abstractmethod
    def produce_single_repr(self, field_data: Union[List[str], str]) -> EmbeddingField:
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError


class Sentence2WordEmbedding(DecombiningInWordsEmbeddingTechnique):
    """
    Class that makes use of a sentence granularity embedding source to produce an embedding matrix with word granularity
    """

    def __init__(self, embedding_source: Union[SentenceEmbeddingLoader, SentenceEmbeddingLearner]):
        # if isinstance(embedding_source, str):
        #     embedding_source = self.from_str_to_embedding_source(embedding_source, SentenceEmbeddingLoader)
        super().__init__(embedding_source)

    def produce_single_repr(self, field_data: Union[List[str], str]) -> EmbeddingField:
        """
        Produces a single matrix where each row is the embedding representation of each token of the sentence,
        while the columns are the hidden dimension of the chosen model

        Args:
            field_data: textual data to complexly represent

        Returns:
            Embedding for each token of the sentence

        """
        field_data = check_not_tokenized(field_data)
        embedding_source: Union[SentenceEmbeddingLoader, SentenceEmbeddingLearner] = self.embedding_source
        words_embeddings = embedding_source.get_embedding_token(field_data)
        return EmbeddingField(words_embeddings)

    def __str__(self):
        return "Sentence2WordEmbedding"

    def __repr__(self):
        return f'Sentence2WordEmbedding(embedding_source={self.embedding_source})'



import unittest

import numpy as np
import torch
from transformers import BertModel, BertTokenizer, T5Model, AutoTokenizer

from clayrs.content_analyzer import Centroid, BertTransformers, T5Transformers
from clayrs.content_analyzer.embeddings import SumStrategy, CatStrategy
from clayrs.content_analyzer.field_content_production_techniques.embedding_technique.combining_technique import \
    SingleToken


class TestBertTransformers(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.tiny_model = BertTransformers('prajjwal1/bert-tiny')

    def test_transformer(self):
        # test that the model loaded is the one passed
        transformers_model = self.tiny_model.model
        self.assertIsInstance(transformers_model, BertModel)

        # test load method
        source = self.tiny_model
        vector_size = source.get_vector_size()

        result = source.load(["this is a phrase", "this is another phrase"])

        self.assertEqual(len(result), 2)
        self.assertEqual(len(result[0]), vector_size)
        self.assertEqual(len(result[1]), vector_size)

    def test_embedding(self):
        # Test embedding obtained with framework and embedding obtained with hugging face library
        sentence = 'This is a beautiful model and very tiny'

        model = BertModel.from_pretrained('prajjwal1/bert-tiny', output_hidden_states=True)
        tokenizer = BertTokenizer.from_pretrained('prajjwal1/bert-tiny')

        encoded = tokenizer.encode_plus(sentence)

        tokens_tensor = torch.tensor([encoded['input_ids']])
        segments_tensors = torch.tensor([encoded['attention_mask']])

        with torch.no_grad():
            model_output = model(tokens_tensor, segments_tensors)
            hidden_states = model_output[2]

        token_embeddings = torch.stack(hidden_states, dim=0)
        token_embeddings = torch.squeeze(token_embeddings, dim=1)
        token_embeddings = token_embeddings.permute(1, 0, 2)

        # sentence embedding from model obtained with cat strategy and centroid
        token_vecs = CatStrategy(1).build_embedding(token_embeddings)
        expected = Centroid().combine(token_vecs)

        # sentence embedding from implemented class
        transformers_model = self.tiny_model
        result = transformers_model.get_embedding(sentence)

        self.assertTrue((expected == result).all())

        # sentence embedding from model obtained with cat strategy and centroid
        token_vecs = SumStrategy(2).build_embedding(token_embeddings)
        expected = SingleToken(0).combine(token_vecs)

        # sentence embedding from implemented class
        transformers_model = BertTransformers('prajjwal1/bert-tiny', vec_strategy=SumStrategy(2),
                                              pooling_strategy=SingleToken(0))
        result = transformers_model.get_embedding(sentence)

        self.assertTrue((expected == result).all())

    def test_embedding_token(self):
        transformers_model = self.tiny_model

        tok_1 = transformers_model.get_embedding_token('Hello, all right.')
        tok_2 = transformers_model.get_embedding_token('Hello, all right.')

        # tokens in same sentence should be equal
        self.assertEqual(tok_1.shape[1], transformers_model.get_vector_size())
        self.assertEqual(tok_2.shape[1], transformers_model.get_vector_size())

        self.assertTrue(np.array_equal(tok_1, tok_2))

        tok_1 = transformers_model.get_embedding_token('Hello how are you?')
        tok_2 = transformers_model.get_embedding_token('Hello, all right.')

        # same token ('Hello') in different sentences should be different
        self.assertEqual(tok_1.shape[1], transformers_model.get_vector_size())
        self.assertEqual(tok_2.shape[1], transformers_model.get_vector_size())

        self.assertFalse(np.array_equal(tok_1[1], tok_2[1]))


class TestT5Transformers(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.tiny_model = T5Transformers('t5-small')

    def test_transformer(self):
        # test that the model loaded is the one passed
        transformers_model = self.tiny_model.model
        self.assertIsInstance(transformers_model, T5Model)

        # test load method
        source = self.tiny_model
        vector_size = source.get_vector_size()

        result = source.load(["this is a phrase", "this is another phrase"])

        self.assertEqual(len(result), 2)
        self.assertEqual(len(result[0]), vector_size)
        self.assertEqual(len(result[1]), vector_size)

    def test_embedding(self):
        # Test embedding obtained with framework and embedding obtained with hugging face library
        sentence = 'This is a beautiful model and very tiny'

        model = T5Model.from_pretrained('t5-small', output_hidden_states=True)
        tokenizer = AutoTokenizer.from_pretrained('t5-small')

        encoded = tokenizer(sentence, truncation=True, return_tensors='pt')

        with torch.no_grad():
            model_output = model.encoder(**encoded)
            hidden_states = model_output['hidden_states']

        token_embeddings = torch.stack(hidden_states, dim=0)
        token_embeddings = torch.squeeze(token_embeddings, dim=1)
        token_embeddings = token_embeddings.permute(1, 0, 2)

        # sentence embedding from model obtained with cat strategy and centroid
        token_vecs = CatStrategy(1).build_embedding(token_embeddings)
        expected = Centroid().combine(token_vecs)

        # sentence embedding from implemented class
        transformers_model = self.tiny_model
        result = transformers_model.get_embedding(sentence)

        self.assertTrue((expected == result).all())

        # sentence embedding from model obtained with cat strategy and centroid
        token_vecs = SumStrategy(2).build_embedding(token_embeddings)
        expected = SingleToken(0).combine(token_vecs)

        # sentence embedding from implemented class
        transformers_model = T5Transformers('t5-small', vec_strategy=SumStrategy(2),
                                            pooling_strategy=SingleToken(0))
        result = transformers_model.get_embedding(sentence)

        self.assertTrue((expected == result).all())

    def test_embedding_token(self):
        transformers_model = self.tiny_model

        tok_1 = transformers_model.get_embedding_token('Hello, all right.')
        tok_2 = transformers_model.get_embedding_token('Hello, all right.')

        # tokens in same sentence should be equal
        self.assertEqual(tok_1.shape[1], transformers_model.get_vector_size())
        self.assertEqual(tok_2.shape[1], transformers_model.get_vector_size())

        self.assertTrue(np.array_equal(tok_1, tok_2))

        tok_1 = transformers_model.get_embedding_token('Hello how are you?')
        tok_2 = transformers_model.get_embedding_token('Hello, all right.')

        # same token ('Hello') in different sentences should be different
        self.assertEqual(tok_1.shape[1], transformers_model.get_vector_size())
        self.assertEqual(tok_2.shape[1], transformers_model.get_vector_size())

        self.assertFalse(np.array_equal(tok_1[1], tok_2[1]))



from __future__ import annotations
import abc
from abc import ABC
from typing import Set, Union, Optional, Generator, Dict, TYPE_CHECKING
import pandas as pd

if TYPE_CHECKING:
    from clayrs.content_analyzer import Ratings

from clayrs.utils.context_managers import get_progbar


class Methodology(ABC):
    """
    Class which, given a *train set* and a *test set*, has the task to calculate which items must be used in
    order to generate a recommendation list

    The methodologies here implemented follow the 'Precision-Oriented Evaluation of Recommender Systems: An Algorithmic
    Comparison' paper
    """

    def __init__(self, only_greater_eq: float = None):
        self._threshold = only_greater_eq

    def _filter_only_greater_eq(self, split_set: Ratings):
        items_list_greater_eq = [interaction.item_id for interaction in split_set if
                                 interaction.score >= self._threshold]

        return items_list_greater_eq

    def filter_all(self, train_set: Ratings, test_set: Ratings,
                   result_as_iter_dict: bool = False) -> Union[pd.DataFrame, Dict[str, Generator]]:
        """
        Concrete method which calculates for all users of the *test set* which items must be used in order to
        generate a recommendation list

        It takes in input a *train set* and a *test set* and returns a single DataFrame or a generator of a python
        dictionary containing, for every user, all items which must be recommended based on the methodology chosen.

        Args:
            train_set: `Ratings` object which contains the train set of every user
            test_set: `Ratings` object which contains the test set of every user
            result_as_iter_dict (bool): If True the output of the method will be a generator of a dictionary that,
                once evaluated, will contains users as a key and list of item that must be predicted as a value.

                    EXAMPLE:
                        `{'u1': ['i1', 'i2', 'i3'], 'u2': ['i1', 'i4'], ...}`
        Returns:
            A DataFrame or a generator of a python dictionary which contains all items which must be recommended to
            every user based on the methodology chosen.
        """
        user_list = set(test_set.user_id_column)

        with get_progbar(user_list) as pbar:
            pbar.set_description(f"Filtering items based on {str(self)}")

            filtered = {user_id: self.filter_single(user_id, train_set, test_set)
                        for user_id in pbar}

        if not result_as_iter_dict:
            generator_expression = ((user_id, item_to_predict)
                                    for user_id, all_items_to_pred in zip(filtered.keys(), filtered.values())
                                    for item_to_predict in set(all_items_to_pred))

            filtered = pd.DataFrame(generator_expression,
                                    columns=['user_id', 'item_id'])

        return filtered

    @abc.abstractmethod
    def filter_single(self, user_id: str, train_set: Ratings, test_set: Ratings) -> Generator:
        """
        Abstract method in which must be specified how to calculate which items must be part of the recommendation list
        of a single user
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __str__(self):
        raise NotImplementedError


class TestRatingsMethodology(Methodology):
    """
    Class which, given a *train set* and a *test set*, has the task to calculate which items must be used in
    order to generate a recommendation list

    With TestRatingsMethodology, given a user $u$, items to recommend for $u$ are simply those items that appear in its
    *test set*

    If the `only_greater_eq` parameter is set, then only items with rating score $>=$ only_greater_eq will be
    returned

    Args:
        only_greater_eq: float which acts as a filter, if specified only items with
            rating score $>=$ only_greater_eq will be returned
    """

    def __init__(self, only_greater_eq: float = None):
        super(TestRatingsMethodology, self).__init__(only_greater_eq)

    def __str__(self):
        return "TestRatingsMethodology"

    def __repr__(self):
        return f"TestRatingsMethodology(only_greater_eq={self._threshold})"

    def filter_single(self, user_id: str, train_set: Ratings, test_set: Ratings) -> Generator:
        """
        Method that returns items that need to be part of the recommendation list of a single user.
        Since it's the TestRatings Methodology, only items that appear in the *test set* of the user will be returned.

        Args:
            user_id: User of which we want to calculate items that must appear in its recommendation list
            train_set: `Ratings` object which contains the train set of every user
            test_set: `Ratings` object which contains the test set of every user
        """
        user_test = test_set.get_user_interactions(user_id)

        if self._threshold is not None:
            filtered_items = (interaction.item_id
                              for interaction in user_test if interaction.score >= self._threshold)
        else:
            # TestRatings just returns the test set of the user
            filtered_items = (interaction.item_id for interaction in user_test)

        return filtered_items


class TestItemsMethodology(Methodology):
    """
    Class which, given a *train set* and a *test set*, has the task to calculate which items must be used in
    order to generate a recommendation list

    With TestItemsMethodology, given a user $u$, items to recommend for $u$ are all items that appear in the
    *test set* of every user excluding those items that appear in the *train set* of $u$

    If the `only_greater_eq` parameter is set, then only items with rating score $>=$ only_greater_eq will be
    returned

    Args:
        only_greater_eq: float which acts as a filter, if specified only items with
            rating score $>=$ only_greater_eq will be returned
    """

    def __init__(self, only_greater_eq: float = None):
        super(TestItemsMethodology, self).__init__(only_greater_eq)

        self._filtered_test_set_items: Optional[Set] = None

    def __str__(self):
        return "TestItemsMethodology"

    def __repr__(self):
        return f"TestItemsMethodology(only_greater_eq={self._threshold})"

    def filter_single(self, user_id: str, train_set: Ratings, test_set: Ratings) -> Generator:
        """
        Method that returns items that need to be part of the recommendation list of a single user.
        Since it's the TestItems Methodology, all items that appear in the *test set* of every user will be returned,
        except for those that appear in the *train set* of the user passed as parameter

        Args:
            user_id: User of which we want to calculate items that must appear in its recommendation list
            train_set: `Ratings` object which contains the train set of every user
            test_set: `Ratings` object which contains the test set of every user
        """
        already_seen_items_it = (interaction.item_id for interaction in train_set.get_user_interactions(user_id))

        filtered_test_set_items = test_set.item_id_column
        if self._threshold is not None:
            if self._filtered_test_set_items is None:
                self._filtered_test_set_items = set(self._filter_only_greater_eq(test_set))

            filtered_test_set_items = self._filtered_test_set_items

        filtered_items = yield from set(filtered_test_set_items) - set(already_seen_items_it)

        return filtered_items


class TrainingItemsMethodology(Methodology):
    """
    Class which, given a *train set* and a *test set*, has the task to calculate which items must be used in
    order to generate a recommendation list

    With TrainingItemsMethodology, given a user $u$, items to recommend for $u$ are all items that appear in the
    'train set' of every user excluding those items that appear in the 'train set' of $u$

    If the `only_greater_eq` parameter is set, then only items with rating score $>=$ only_greater_eq will be
    returned

    Args:
        only_greater_eq: float which acts as a filter, if specified only items with
            rating score $>=$ only_greater_eq will be returned
    """

    def __init__(self, only_greater_eq: float = None):
        super(TrainingItemsMethodology, self).__init__(only_greater_eq)

        self._filtered_train_set_items: Optional[Set] = None

    def __str__(self):
        return "TrainingItemsMethodology"

    def __repr__(self):
        return f"TrainingItemsMethodology(only_greater_eq={self._threshold})"

    def filter_single(self, user_id: str, train_set: Ratings, test_set: Ratings) -> Generator:
        """
        Method that returns items that needs to be part of the recommendation list of a single user.
        Since it's the TrainingItems Methodology, all items that appear in the *train set* of every user will be
        returned, except for those that appear in the *train set* of the user passed as parameter

        Args:
            user_id: User of which we want to calculate items that must appear in its recommendation list
            train_set: `Ratings` object which contains the train set of every user
            test_set: `Ratings` object which contains the test set of every user
        """
        already_seen_items_it = (interaction.item_id for interaction in train_set.get_user_interactions(user_id))

        filtered_train_set_items = train_set.item_id_column
        if self._threshold is not None:
            if self._filtered_train_set_items is None:
                self._filtered_train_set_items = set(self._filter_only_greater_eq(train_set))

            filtered_train_set_items = self._filtered_train_set_items

        filtered_items = yield from set(filtered_train_set_items) - set(already_seen_items_it)

        return filtered_items


class AllItemsMethodology(Methodology):
    """
    Class which, given a *train set* and a *test set*, has the task to calculate which items must be used in
    order to generate a recommendation list

    With AllItemsMethodology, given a user $u$, items to recommend for $u$ are all items that appear in 'items_list'
    parameter excluding those items that appear in the *train set* of $u$

    Args:
        items_list: Items set that must appear in the recommendation list of every user
    """

    def __init__(self, items_list: Set[str]):
        self._items_list = items_list
        super(AllItemsMethodology, self).__init__(None)

    def __str__(self):
        return "AllItemsMethodology"

    def __repr__(self):
        return f"AllItemsMethodology(items_list={self._items_list})"

    def filter_single(self, user_id: str, train_set: Ratings, test_set: Ratings) -> Generator:
        """
        Method that returns items that needs to be part of the recommendation list of a single user.
        Since it's the AllItems Methodology, all items that appear in the `items_list` parameter of the constructor
        will be returned, except for those that appear in the *train set* of the user passed as parameter

        Args:
            user_id: User of which we want to calculate items that must appear in its recommendation list
            train_set: `Ratings` object which contains the train set of every user
            test_set: `Ratings` object which contains the test set of every user
        """
        already_seen_items_it = (interaction.item_id for interaction in train_set.get_user_interactions(user_id))

        filtered_items = yield from set(self._items_list) - set(already_seen_items_it)

        return filtered_items
